# Generated by Django 3.1.5 on 2021-01-28 17:09

import datetime

import colorfield.fields
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models

import wkz.models


class Migration(migrations.Migration):

    replaces = [
        ("wkz", "0002_auto_20200704_1728"),
        ("wkz", "0003_auto_20200709_0742"),
        ("wkz", "0004_traces_distance_list"),
        ("wkz", "0005_auto_20201010_1101"),
        ("wkz", "0006_settings_run_file_importer_in_background"),
        ("wkz", "0007_auto_20201229_1033"),
        ("wkz", "0008_bestsection_bestsectiontype"),
        ("wkz", "0009_auto_20201230_0827"),
        ("wkz", "0010_auto_20201230_1038"),
        ("wkz", "0011_bestsection_section_distance"),
        ("wkz", "0012_bestsectiontopscores"),
        ("wkz", "0013_auto_20210101_1818"),
        ("wkz", "0014_auto_20210112_1924"),
    ]

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Sport",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=24, unique=True, verbose_name="Sport Name:")),
                ("color", colorfield.fields.ColorField(default="#42FF71", max_length=18, verbose_name="Color:")),
                ("icon", models.CharField(max_length=24, verbose_name="Icon:")),
                ("slug", models.SlugField(blank=True, max_length=100, unique=True)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Traces",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("path_to_file", models.CharField(max_length=200)),
                ("file_name", models.CharField(editable=False, max_length=100)),
                ("md5sum", models.CharField(max_length=32, unique=True)),
                ("altitude_list", models.CharField(default="[]", max_length=10000000000)),
                ("heart_rate_list", models.CharField(default="[]", max_length=10000000000)),
                ("max_altitude", models.FloatField(blank=True, null=True)),
                ("min_altitude", models.FloatField(blank=True, null=True)),
                ("aerobic_training_effect", models.FloatField(blank=True, null=True)),
                ("anaerobic_training_effect", models.FloatField(blank=True, null=True)),
                ("avg_cadence", models.IntegerField(blank=True, null=True)),
                ("avg_heart_rate", models.IntegerField(blank=True, null=True)),
                ("avg_speed", models.FloatField(blank=True, null=True)),
                ("avg_temperature", models.IntegerField(blank=True, null=True)),
                ("cadence_list", models.CharField(default="[]", max_length=10000000000)),
                ("max_cadence", models.IntegerField(blank=True, null=True)),
                ("max_heart_rate", models.IntegerField(blank=True, null=True)),
                ("max_speed", models.FloatField(blank=True, null=True)),
                ("max_temperature", models.IntegerField(blank=True, null=True)),
                ("speed_list", models.CharField(default="[]", max_length=10000000000)),
                ("temperature_list", models.CharField(default="[]", max_length=10000000000)),
                ("min_cadence", models.IntegerField(blank=True, null=True)),
                ("min_heart_rate", models.IntegerField(blank=True, null=True)),
                ("min_speed", models.FloatField(blank=True, null=True)),
                ("min_temperature", models.IntegerField(blank=True, null=True)),
                ("calories", models.IntegerField(blank=True, null=True)),
                ("timestamps_list", models.CharField(default="[]", max_length=10000000000)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("distance_list", models.CharField(default="[]", max_length=10000000000)),
                ("latitude_list", models.CharField(default="[]", max_length=10000000000)),
                ("longitude_list", models.CharField(default="[]", max_length=10000000000)),
            ],
        ),
        migrations.CreateModel(
            name="Lap",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("start_time", models.DateTimeField()),
                ("end_time", models.DateTimeField()),
                ("elapsed_time", models.DurationField()),
                ("start_lat", models.FloatField(blank=True, null=True)),
                ("start_long", models.FloatField(blank=True, null=True)),
                ("end_lat", models.FloatField(blank=True, null=True)),
                ("end_long", models.FloatField(blank=True, null=True)),
                ("distance", models.FloatField(blank=True, null=True)),
                ("speed", models.FloatField(blank=True, null=True)),
                ("trace", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wkz.traces")),
                ("trigger", models.CharField(max_length=120)),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("label", models.CharField(blank=True, max_length=100, null=True, verbose_name="Label:")),
            ],
        ),
        migrations.CreateModel(
            name="Activity",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(default="unknown", max_length=200, verbose_name="Activity Name:")),
                ("date", models.DateTimeField(default=django.utils.timezone.now, verbose_name="Date:")),
                ("duration", models.DurationField(default=datetime.timedelta(seconds=1800), verbose_name="Duration:")),
                ("distance", models.FloatField(blank=True, default=0, null=True, verbose_name="Distance:")),
                ("description", models.CharField(blank=True, max_length=600, null=True, verbose_name="Description:")),
                (
                    "sport",
                    models.ForeignKey(
                        default=wkz.models.default_sport,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        to="wkz.sport",
                        verbose_name="Sport:",
                    ),
                ),
                (
                    "trace_file",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="wkz.traces"
                    ),
                ),
                ("is_demo_activity", models.BooleanField(default=False, verbose_name="Is this a Demo Activity:")),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Settings",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "path_to_trace_dir",
                    models.CharField(
                        default="/workspaces/workoutizer/tracks",
                        max_length=120,
                        verbose_name="Path to Traces Directory:",
                    ),
                ),
                (
                    "path_to_garmin_device",
                    models.CharField(
                        default="/run/user/1000/gvfs/", max_length=120, verbose_name="Path to Garmin Device:"
                    ),
                ),
                (
                    "number_of_days",
                    models.IntegerField(
                        choices=[(9999, "all"), (365, 365), (180, 180), (90, 90), (30, 30), (10, 10)], default=30
                    ),
                ),
                (
                    "delete_files_after_import",
                    models.BooleanField(default=False, verbose_name="Delete fit Files after Copying: "),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="BestSection",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("start_index", models.IntegerField()),
                ("end_index", models.IntegerField()),
                ("max_value", models.FloatField()),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("activity", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wkz.activity")),
                ("section_type", models.CharField(max_length=120)),
                ("section_distance", models.IntegerField(default=0)),
            ],
        ),
        migrations.CreateModel(
            name="BestSectionTopScores",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("rank", models.IntegerField(choices=[(1, "First"), (2, "Second"), (3, "Third")])),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("activity", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wkz.activity")),
                ("sport", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="wkz.sport")),
                (
                    "section",
                    models.ForeignKey(default=1, on_delete=django.db.models.deletion.CASCADE, to="wkz.bestsection"),
                ),
            ],
        ),
    ]



import math

import pytest

from wkz.gis.geo import (
    add_elevation_data_to_coordinates,
    calculate_distance_between_points,
    get_list_of_coordinates,
    get_location_name,
    get_total_distance_of_trace,
)


def test_calculate_distance_between_points__same_points():
    coordinate_1 = (48.0, 9.0)
    coordinate_2 = (48.0, 9.0)
    assert calculate_distance_between_points(coordinate_1, coordinate_2) == 0.0


def test_calculate_distance_between_points__different_points():
    coordinate_1 = (48.123, 9.456)
    coordinate_2 = (49.678, 9.567)
    distance = calculate_distance_between_points(coordinate_1, coordinate_2)
    assert math.isclose(distance, 173291.21920642233, abs_tol=0.01)


def test_calculate_distance_between_points__other_points():
    coordinate_1 = (48.0, 8.0)
    coordinate_2 = (48.0, 8.1)
    distance = calculate_distance_between_points(coordinate_1, coordinate_2)
    assert math.isclose(distance, 7448.684105664539, abs_tol=0.01)


def test_get_total_distance_of_trace__basic():
    with pytest.raises(ValueError):
        get_total_distance_of_trace(longitude_list=[0], latitude_list=[0])
    with pytest.raises(ValueError):
        get_total_distance_of_trace(longitude_list=[0], latitude_list=[0, 1])
    with pytest.raises(ValueError):
        get_total_distance_of_trace(longitude_list=[0, 1], latitude_list=[0])
    assert get_total_distance_of_trace(longitude_list=[48, 48], latitude_list=[8, 9]) == 111.32
    assert get_total_distance_of_trace(longitude_list=[48, 48], latitude_list=[-8, -9]) == 111.32
    assert get_total_distance_of_trace(longitude_list=[-48, -48], latitude_list=[-8, -9]) == 111.32
    assert get_total_distance_of_trace(longitude_list=[-48, -48], latitude_list=[8, 9]) == 111.32


def test_get_total_distance_of_trace__reversed_points():
    assert get_total_distance_of_trace(longitude_list=[99, 98], latitude_list=[16, 16]) == 107.01


def test_add_elevation_data_to_coordinates():
    assert (
        add_elevation_data_to_coordinates(
            coordinates=[(8, 49), (9, 50), (10, 51)],
            altitude=[248, 249, 250],
        )
        == [(8, 49, 248), (9, 50, 249), (10, 51, 250)]
    )


def test__get_location_name():
    coordinate = (48.1234, 8.9123)
    location_name = get_location_name(coordinate=coordinate)
    assert location_name == "Heidenstadt"

    coordinate = (49.47950, 8.47102)
    location_name = get_location_name(coordinate=coordinate)
    assert location_name == "Mannheim"

    # this would raise an exception in geopy so we expect to get None
    coordinate = (-90, 90)
    location_name = get_location_name(coordinate=coordinate)
    assert location_name is None


def test_get_other_location_names():
    coordinate = (49.46278064511717, 8.160513974726202)
    location_name = get_location_name(coordinate=coordinate)
    assert location_name == "Bad Dürkheim"


def test_get_list_of_coordinates__empty_data():
    lon = []
    lat = []
    coordinates = get_list_of_coordinates(lon, lat)

    assert coordinates == []


def test_get_list_of_coordinates__dummy_data():
    lon = [48.1, 48.2, 48.3, 48.4, 48.5, 48.6, 48.7, 48.8, 48.9, 49.0, 49.1]
    lat = [9.3, 9.4, 9.5, 9.6, 9.7, 9.8, 9.9, 10.0, 10.1, 10.2, 10.3]
    coordinates = get_list_of_coordinates(lon, lat)

    # note that because of configuration.every_nth_value only every nth value will be used
    assert coordinates == [(48.1, 9.3), (48.6, 9.8), (49.1, 10.3)]


def test_get_list_of_coordinates__data_with_gaps():
    lon = [None, 48.1234, None, 49.2345, None, 49.2345]
    lat = [None, 9.4567, None, 10.5678, None, 11.2345]
    coordinates = get_list_of_coordinates(lon, lat)

    # note that because of configuration.every_nth_value only every nth value will be used
    assert coordinates == [(49.2345, 11.2345), (49.2345, 11.2345)]



import os
from pathlib import Path

from huey import SqliteHuey

from workoutizer.logger import get_logging_config

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
MEDIA_ROOT = os.path.join(BASE_DIR, "media")
INITIAL_TRACE_DATA_DIR = os.path.join(BASE_DIR, "setup", "initial_trace_data")

if os.getenv("WKZ_ENV", None) == "devel":
    WORKOUTIZER_DIR = BASE_DIR
else:
    USER_HOME = Path.home()
    WORKOUTIZER_DIR = os.path.join(str(USER_HOME), ".wkz")
SQLITE_FILE = "db.sqlite3"
WORKOUTIZER_DB_PATH = os.path.join(WORKOUTIZER_DIR, SQLITE_FILE)
TRACKS_DIR = os.path.join(WORKOUTIZER_DIR, "tracks")

STATIC_ROOT = os.path.join(BASE_DIR, "static")
STATIC_URL = "/static/"

# SECURITY WARNING: keep the secret key used in production secret!
# however, as long as workoutizer is only used locally this is not an issue
SECRET_KEY = "h#ppx^(%ya18qrm+hgzf-vxr^t=r57k_65_hr73f^-n)@qc9as"

DEBUG = False

ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "wkz",
    "colorfield",
    "rest_framework",
    "channels",
    "django_eventstream",
    "huey.contrib.djhuey",
]

MIDDLEWARE = [
    "django_grip.GripMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "workoutizer.urls"
MESSAGE_STORAGE = "django.contrib.messages.storage.cookie.CookieStorage"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "workoutizer.wsgi.application"
ASGI_APPLICATION = "workoutizer.asgi.application"

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": WORKOUTIZER_DB_PATH,
    }
}

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

LANGUAGE_CODE = "en-us"
TIME_ZONE = "Europe/Berlin"
USE_I18N = True
USE_L10N = True
USE_TZ = True

PLOT_WIDTH = 1110
PLOT_HEIGHT = 300

REST_FRAMEWORK = {
    "DEFAULT_PERMISSION_CLASSES": [
        "rest_framework.permissions.AllowAny",
    ],
}

# set auto primary key to BigAutoField explicitly to suppress warning
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# plotting
trace_line_width = 3.5
trace_line_opacity = 0.9

LOGGING = get_logging_config(
    django_log_level=os.getenv("DJANGO_LOG_LEVEL", "WARNING"),
    wkz_log_level=os.getenv("WKZ_LOG_LEVEL", "INFO"),
    huey_log_level=os.getenv("HUEY_LOG_LEVEL", "WARNING"),
    path_to_log_dir=WORKOUTIZER_DIR,
)

HUEY = SqliteHuey(filename="/tmp/demo.db")



import datetime

from wkz.file_helper.gpx_exporter import _build_gpx, _fill_list_of_timestamps


def test__build_gpx(trace_coordinates, gpx_string):
    list_of_timestamps = _fill_list_of_timestamps(
        start=datetime.datetime(2019, 7, 12), duration=datetime.timedelta(minutes=4), length=len(trace_coordinates)
    )
    assert (
        _build_gpx(
            time=datetime.datetime(2019, 7, 12),
            file_name="test",
            coordinates=trace_coordinates,
            timestamps=list_of_timestamps,
            sport="Running",
        )
        == gpx_string
    )


def test__build_gpx_with_elevation(trace_coordinates_with_elevation, gpx_string_with_elevation):
    list_of_timestamps = _fill_list_of_timestamps(
        start=datetime.datetime(2019, 7, 12),
        duration=datetime.timedelta(minutes=4),
        length=len(trace_coordinates_with_elevation),
    )
    assert (
        _build_gpx(
            time=datetime.datetime(2019, 7, 12),
            file_name="test",
            coordinates=trace_coordinates_with_elevation,
            timestamps=list_of_timestamps,
            sport="Running",
        )
        == gpx_string_with_elevation
    )


def test__fill_list_of_timestamps():
    length = 3
    assert (
        len(
            _fill_list_of_timestamps(
                start=datetime.datetime(2019, 7, 12),
                duration=datetime.timedelta(minutes=30),
                length=length,
            )
        )
        == length
    )
    assert (
        _fill_list_of_timestamps(
            start=datetime.datetime(2019, 7, 12),
            duration=datetime.timedelta(minutes=30),
            length=length,
        )
        == ["2019-07-12T12:00:00Z", "2019-07-12T12:10:00Z", "2019-07-12T12:20:00Z"]
    )



from wkz import models


def test_traces__creation_of_file_name(db):
    trace = models.Traces(path_to_file="/some/dummy/path/to/file.gpx", md5sum="asdfasdf")
    trace.save()
    assert trace.file_name == "file.gpx"


def test_sport__slugify_of_sport_name(db):
    sport = models.Sport(name="Some Crazy Stuff", color="red", icon="Bike")
    sport.save()
    assert sport.slug == "some-crazy-stuff"


def test_activity__with_default_unknown_sport(db):
    name = "My Activity 1"
    activity_1 = models.Activity(name=name)
    activity_1.save()
    assert activity_1.name == name
    assert activity_1.sport.name == "unknown"
    assert activity_1.sport.icon == "question-circle"
    assert activity_1.sport.color == "gray"
    # add second activity to verify that the same unknown sport instance is used
    name = "My Activity 2"
    activity_2 = models.Activity(name=name)
    activity_2.save()
    assert activity_2.name == name
    assert activity_2.sport == activity_1.sport



import logging

import pandas as pd
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.forms.models import model_to_dict
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse
from django.views.generic import DeleteView, View

from wkz import configuration, models
from wkz.forms import AddSportsForm
from wkz.plotting.plot_history import plot_history
from wkz.tools.utils import remove_microseconds
from wkz.views import (
    MapView,
    PlotView,
    get_all_form_field_ids,
    get_flat_list_of_pks_of_activities_in_top_awards,
    get_summary_of_all_activities,
)

log = logging.getLogger(__name__)


class SportsView(View):
    template_name = "sport/sports.html"

    def get(self, request):
        sports = models.Sport.objects.all().order_by("name")
        sport_data = {}
        for sport in sports:
            if sport.name not in configuration.protected_sports:
                activities_df = pd.DataFrame(
                    list(models.Activity.objects.filter(sport=sport.id).values("duration", "distance"))
                )
                setattr(sport, "total_count", len(activities_df))
                if activities_df.empty:
                    setattr(sport, "total_distance", 0)
                    setattr(sport, "total_duration", 0)
                else:
                    setattr(sport, "total_distance", round(activities_df["distance"].sum(), 2))
                    setattr(sport, "total_duration", remove_microseconds(activities_df["duration"].sum()))
        return render(
            request,
            self.template_name,
            {"sports": sports, "page_name": "Sports", "form_field_ids": get_all_form_field_ids(), **sport_data},
        )


class SportView(MapView, PlotView):
    template_name = "sport/sport.html"

    def get(self, request, sports_name_slug):
        log.debug(f"got sports name: {sports_name_slug}")
        settings = models.get_settings()
        if sports_name_slug == "undefined":
            log.warning("could not find sport - redirecting to home")
            return HttpResponseRedirect(reverse("home"))
        sport = models.Sport.objects.get(slug=sports_name_slug)
        activities = self.get_activity_data_for_plots(sport_id=sport.id)
        context = {}
        sports = models.Sport.objects.all().order_by("name")
        summary = get_summary_of_all_activities(sport_slug=sports_name_slug)
        if activities:
            script_history, div_history = plot_history(
                activities=activities,
                sport_model=models.Sport,
                number_of_days=settings.number_of_days,
            )
            context["script_history"] = script_history
            context["div_history"] = div_history
            context["activities_selected_for_plot"] = True
        else:
            context["activities_selected_for_plot"] = False
        map_context = super(SportView, self).get(request=request, list_of_activities=activities)
        if sport.evaluates_for_awards:
            top_awards = get_flat_list_of_pks_of_activities_in_top_awards(sports_name_slug)
            context["top_awards"] = top_awards
        try:
            sport = model_to_dict(models.Sport.objects.get(slug=sports_name_slug))
            sport["slug"] = sports_name_slug
        except ObjectDoesNotExist:
            log.critical("this sport does not exist")
            raise Http404
        page = 0
        return render(
            request,
            self.template_name,
            {
                **map_context,
                **context,
                "page_name": sport["name"],
                "available_sport_names": [sport.name for sport in sports],
                "current_page": page,
                "is_last_page": False,
                "sports": sports,
                "summary": summary,
                "sport": sport,
                "form_field_ids": get_all_form_field_ids(),
            },
        )


def add_sport_view(request):
    sports = models.Sport.objects.all().order_by("name")
    if request.method == "POST":
        form = AddSportsForm(request.POST)
        if form.is_valid():
            instance = form.save()
            instance.save()
            messages.success(request, f"Successfully added '{form.cleaned_data['name']}'")
            return HttpResponseRedirect(reverse("sports"))
        else:
            log.warning(f"form invalid: {form.errors}")
    else:
        form = AddSportsForm()
    return render(
        request,
        "sport/add_sport.html",
        {"sports": sports, "form": form, "page_name": "Add Sport", "form_field_ids": get_all_form_field_ids()},
    )


def edit_sport_view(request, sports_name_slug):
    sports = models.Sport.objects.all().order_by("name")
    sport = models.Sport.objects.get(slug=sports_name_slug)
    if sport.name in configuration.protected_sports:
        messages.warning(request, f"Can't edit sport '{sport.name}'")
        return HttpResponseRedirect(f"/sport/{sport.slug}")
    form = AddSportsForm(request.POST or None, instance=sport)
    if request.method == "POST":
        if form.is_valid():
            log.debug(f"got valid form: {form.cleaned_data}")
            form.save()
            messages.success(request, f"Successfully modified '{form.cleaned_data['name']}' Sport")
            return HttpResponseRedirect(f"/sport/{sport.slug}/edit/")
        else:
            log.warning(f"form invalid: {form.errors}")
    return render(
        request,
        "sport/edit_sport.html",
        {
            "sports": sports,
            "sport": sport,
            "form": form,
            "form_field_ids": get_all_form_field_ids(),
            "page_name": f"Edit Sport: {sport.name}",
        },
    )


class SportDeleteView(DeleteView):
    template_name = "sport/sport_confirm_delete.html"
    model = models.Sport
    slug_field = "slug"
    success_url = "/sports/"

    def get(self, request, *args, **kwargs):
        sports = models.Sport.objects.all().order_by("name")
        sport = models.Sport.objects.get(slug=kwargs["slug"])
        if sport.name in configuration.protected_sports:
            messages.warning(request, f"Can't delete sport '{sport.name}'")
            return HttpResponseRedirect(f"/sport/{sport.slug}")
        return render(
            request,
            self.template_name,
            {
                "sports": sports,
                "sport": sport,
                "form_field_ids": get_all_form_field_ids(),
                "page_name": f"Delete Sport: {sport.name}",
            },
        )



from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    reraise=True,
    wait=wait_exponential(multiplier=3, min=1, max=5),
    stop=stop_after_attempt(3),
)
def delayed_assertion(func, operator, right):
    """
    Helper function to check if a function evaluates to given value with a fixed number of retries.
    """
    assert operator(func(), right)



# Generated by Django 3.1.6 on 2021-03-03 05:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wkz", "0003_auto_20210218_2232"),
    ]

    operations = [
        migrations.AddField(
            model_name="settings",
            name="path_to_activities_on_device",
            field=models.CharField(
                default="/Primary/GARMIN/Activity/",
                max_length=120,
                verbose_name="Path to Activities Folder on Garmin Device:",
            ),
        ),
    ]



import os
import subprocess
import sys
from pathlib import Path

import click
import luddite
import requests
from django.core.management import execute_from_command_line
from django.db.utils import OperationalError

from wkz.tools.utils import get_local_ip_address
from workoutizer import __version__
from workoutizer.settings import TRACKS_DIR, WORKOUTIZER_DB_PATH, WORKOUTIZER_DIR

os.environ["DJANGO_SETTINGS_MODULE"] = "workoutizer.settings"


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(__version__)
    ctx.exit()


@click.option(
    "-v", "--version", help="prints the version", is_flag=True, callback=print_version, expose_value=False, is_eager=True
)
@click.group()
def wkz():
    pass


@click.option("-d", "--demo", help="adds demo activity data", is_flag=True)
@click.command(
    help="Mandatory command to initialize workoutizer. This fetches the static files, creates the database, "
    "applies the required migrations and inserts the demo activities."
)
def init(demo):
    _init(import_demo_activities=demo)


def _is_main_run():
    return os.environ.get("RUN_MAIN", None) != "true"


@click.argument("url", default="")
@click.command(
    help="Run workoutizer. Passing the local ip address and port is optionally. In case of no ip address "
    "being passed, it will be determined automatically. Usage, e.g.: 'wkz run 0.0.0.0:8000'."
)
def run(url):
    if not url:
        if os.getenv("WKZ_ENV") == "devel":
            url = "0.0.0.0:8000"
        else:
            url = f"{get_local_ip_address()}:8000"
    if _is_main_run():
        click.echo(f"using workoutizer home at {WORKOUTIZER_DIR}")
    with HueyManager():
        execute_from_command_line(["manage.py", "runserver", url, "--insecure"])


@click.argument("cmd", nargs=1)
@click.command(
    help="Pass commands to django's manage.py. Convenience function to access all django commands. Usage, e.g.: "
    "wkz manage 'runserver 0.0.0.0:8000 --noreload'"
)
def manage(cmd):
    execute_from_command_line(["manage.py"] + cmd.split(" "))


@click.command(help="Check for a newer version and install if there is any.")
def upgrade():
    _upgrade()


@click.command(help="Stop a running workoutizer instance.")
def stop():
    _stop()


@click.command(help="Check if workoutizer was properly installed")
def check():
    _check()


@click.command(help="Reimport all activities to update the given data.")
def reimport():
    _reimport()


wkz.add_command(upgrade)
wkz.add_command(stop)
wkz.add_command(init)
wkz.add_command(run)
wkz.add_command(manage)
wkz.add_command(check)
wkz.add_command(reimport)


def _upgrade():

    latest_version = luddite.get_version_pypi("workoutizer")
    from workoutizer import __version__ as current_version

    if latest_version == current_version:
        click.echo(f"No update available. You are running the latest version: {latest_version}")
    else:
        click.echo(f"found newer version: {latest_version}, you have {current_version} installed - upgrading...")
        _pip_install("workoutizer", upgrade=True)
        execute_from_command_line(["manage.py", "collectstatic", "--noinput"])
        execute_from_command_line(["manage.py", "migrate"])
        execute_from_command_line(["manage.py", "check"])
        click.echo(f"Successfully upgraded from {current_version} to {latest_version}")


def _build_home() -> None:
    if Path(WORKOUTIZER_DIR).is_dir():
        if Path(TRACKS_DIR).is_dir():
            # both folders are already created - do nothing
            return
        else:
            Path(TRACKS_DIR).mkdir(exist_ok=True)
    else:
        Path(WORKOUTIZER_DIR).mkdir(exist_ok=True)
        Path(TRACKS_DIR).mkdir(exist_ok=True)


def _init(import_demo_activities=False):
    _build_home()
    if Path(WORKOUTIZER_DB_PATH).is_file():
        execute_from_command_line(["manage.py", "check"])
        from wkz import models

        try:
            if models.Settings.objects.count() == 1:
                if models.Activity.objects.count() == 0 and import_demo_activities:
                    click.echo("Found initialized db, but with no demo activity, importing...")
                else:
                    click.echo(
                        f"Found initialized db at {WORKOUTIZER_DB_PATH}, maybe you want to run wkz \n"
                        "instead. If you really want to initialize wkz consider removing the existing db file. \n"
                        "Aborting."
                    )
                    return
        except OperationalError:
            pass  # means required tables are not set up - continuing with applying migrations
    execute_from_command_line(["manage.py", "collectstatic", "--noinput"])
    execute_from_command_line(["manage.py", "migrate"])
    from wkz import models

    # insert settings
    models.get_settings()
    _check()

    if import_demo_activities:
        # import demo activities
        from wkz.demo import prepare_import_of_demo_activities
        from wkz.file_importer import run_importer__dask

        prepare_import_of_demo_activities(models)
        run_importer__dask(models, importing_demo_data=True)
        click.echo(f"Database and track files are stored in: {WORKOUTIZER_DIR}")


def _pip_install(package, upgrade: bool = False):
    if upgrade:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package, "--upgrade"])
    else:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package])


def _stop():
    host = get_local_ip_address()
    click.echo(f"stopping workoutizer at {host}")
    url = f"http://{host}:8000/stop/"
    try:
        requests.post(url)
        click.echo("Stopped.")
    except requests.exceptions.ConnectionError:
        click.echo("Workoutizer is not running.")


class NotInitializedError(Exception):
    pass


def _check():
    try:
        execute_from_command_line(["manage.py", "check"])

        # ensure that at least settings are present
        from wkz import models

        if models.Settings.objects.count() != 1:
            raise NotInitializedError("ERROR: Make sure to execute 'wkz init' first")

    except OperationalError:
        raise NotInitializedError("ERROR: Make sure to execute 'wkz init' first")


def _reimport():
    _check()

    from wkz import models
    from wkz.file_importer import run_importer__dask

    run_importer__dask(models, reimporting=True)


class HueyManager:
    def __init__(self):
        self.process = None

    def __enter__(self):
        if _is_main_run():
            self.process = subprocess.Popen([sys.executable, "-m", "django", "run_huey"])
            return self.process

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.process:
            self.process.terminate()



# Generated by Django 3.1.6 on 2021-03-21 17:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wkz", "0005_remove_settings_path_to_activities_on_device"),
    ]

    operations = [
        migrations.RenameField(
            model_name="bestsection",
            old_name="section_distance",
            new_name="distance",
        ),
        migrations.RenameField(
            model_name="bestsection",
            old_name="end_index",
            new_name="end",
        ),
        migrations.RenameField(
            model_name="bestsection",
            old_name="section_type",
            new_name="kind",
        ),
        migrations.RenameField(
            model_name="bestsection",
            old_name="start_index",
            new_name="start",
        ),
        migrations.AlterField(
            model_name="settings",
            name="number_of_days",
            field=models.IntegerField(
                choices=[
                    (9999, "all"),
                    (1095, "3 years"),
                    (730, "2 years"),
                    (365, "1 year"),
                    (180, "180 days"),
                    (90, "90 days"),
                    (30, "30 days"),
                    (10, "10 days"),
                ],
                default=30,
            ),
        ),
    ]



import datetime
import json
import shutil
from pathlib import Path

import pytz

from wkz import models
from wkz.best_sections.generic import activity_suitable_for_awards
from wkz.demo import copy_demo_fit_files_to_track_dir
from wkz.file_importer import run_importer__dask
from wkz.tools.utils import calc_md5
from workoutizer import settings as django_settings


def test_activity_data_in_db_after_import(import_one_activity):
    """
    This test corresponds to tests.unit_tests.file_helper.test_fit_parser.test__parse_records
    but checks the integration with the database and the django model fields.
    """

    assert models.Activity.objects.count() == 0
    import_one_activity("example.fit")
    assert models.Activity.objects.count() == 1

    a = models.Activity.objects.get()

    assert a.sport.slug == "unknown"
    assert a.distance == 5.84
    assert a.duration == datetime.timedelta(seconds=3164)
    assert a.date == datetime.datetime(2019, 9, 14, 15, 22, tzinfo=pytz.UTC)
    assert a.trace_file.avg_heart_rate == 130
    assert a.trace_file.avg_speed == 1.845
    assert a.trace_file.avg_cadence == 64
    assert a.trace_file.avg_temperature == 27
    assert a.trace_file.calories == 432
    assert a.trace_file.aerobic_training_effect == 2.7
    assert a.trace_file.anaerobic_training_effect == 0.3
    # check lengths of list attributes
    assert len(json.loads(a.trace_file.heart_rate_list)) == 1224
    assert len(json.loads(a.trace_file.altitude_list)) == 1224
    assert len(json.loads(a.trace_file.latitude_list)) == 1224
    assert len(json.loads(a.trace_file.longitude_list)) == 1224
    assert len(json.loads(a.trace_file.distance_list)) == 1224
    assert len(json.loads(a.trace_file.cadence_list)) == 1224
    assert len(json.loads(a.trace_file.temperature_list)) == 1224
    assert len(json.loads(a.trace_file.speed_list)) == 1224
    assert len(json.loads(a.trace_file.timestamps_list)) == 1224
    # sanity check to see if element in list attributes
    assert 1.605 in json.loads(a.trace_file.speed_list)
    assert 8.697221484035255 in json.loads(a.trace_file.longitude_list)
    assert 49.40601873211563 in json.loads(a.trace_file.latitude_list)
    assert 1.6 in json.loads(a.trace_file.distance_list)
    assert 248.9 in json.loads(a.trace_file.altitude_list)
    assert 99 in json.loads(a.trace_file.heart_rate_list)
    assert 61 in json.loads(a.trace_file.cadence_list)
    assert 31 in json.loads(a.trace_file.temperature_list)
    assert 1568467325.0 in json.loads(a.trace_file.timestamps_list)
    # check laps
    laps = models.Lap.objects.filter(trace=a.trace_file)
    assert len(laps) == 7
    lap = laps[3]
    assert lap.start_time == datetime.datetime(2019, 9, 14, 14, 56, 48, tzinfo=pytz.UTC)
    assert lap.end_time == datetime.datetime(2019, 9, 14, 15, 3, 23, tzinfo=pytz.UTC)
    assert lap.distance == 1000.0
    assert lap.start_lat == 49.39987546764315
    assert lap.start_long == 8.704907605424525
    assert lap.end_lat == 49.405560996383436
    assert lap.end_long == 8.702977923676373
    assert lap.trigger == "distance"
    assert lap.speed == 2.52
    assert lap.label is None
    # check total ascent/descent
    assert a.trace_file.total_ascent == 234
    assert a.trace_file.total_descent == 233


def test_import_of_demo_activities(import_demo_data, client):
    # verify activities got imported
    all_activities = models.Activity.objects.all()
    assert len(all_activities) == 19

    for activity in all_activities:
        response = client.get(f"/activity/{activity.pk}")
        assert response.status_code == 200

    swimming = models.Activity.objects.filter(sport__slug="swimming")
    assert len(swimming) == 9
    # verify that swimming activities are considered for awards
    for s in swimming:
        assert s.evaluates_for_awards is True

    jogging = models.Activity.objects.filter(sport__slug="jogging")
    assert len(jogging) == 4
    # verify that jogging activities are considered for awards
    for j in jogging:
        assert j.evaluates_for_awards is True

    cycling = models.Activity.objects.filter(sport__slug="cycling")
    assert len(cycling) == 3
    # verify that cycling activities are considered for awards
    for c in cycling:
        assert c.evaluates_for_awards is True

    hiking = models.Activity.objects.filter(sport__slug="hiking")
    assert len(hiking) == 3

    # verify that hiking activities are not considered for awards
    # the entire hiking sport is set to False
    sport = models.Sport.objects.get(slug="hiking")
    assert sport.evaluates_for_awards is False
    for h in hiking:
        # each individual activity is set to True
        assert h.evaluates_for_awards is True
        # and thus at the end the activity evaluates to False
        assert activity_suitable_for_awards(h) is False

    # verify that best sections got parsed and imported
    best_sections_cnt = models.BestSection.objects.count()
    assert best_sections_cnt > 20

    # number of best sections is the sum of fastest and climb sections
    fastest_sections_cnt = models.BestSection.objects.filter(kind="fastest").count()
    climb_sections_cnt = models.BestSection.objects.filter(kind="climb").count()
    assert best_sections_cnt == fastest_sections_cnt + climb_sections_cnt


def test_import_of_activities__not_evaluates_for_awards(import_one_activity):
    # insert default sport
    sport = models.default_sport()

    assert models.Sport.objects.count() == 1
    assert models.Settings.objects.count() == 1

    # get default sport
    sport = models.Sport.objects.get()
    assert sport.name == "unknown"
    assert sport.evaluates_for_awards is True

    # change flag to false
    sport.evaluates_for_awards = False
    sport.save()
    assert sport.evaluates_for_awards is False

    # now import activity and verify that best sections still got saved to the db
    assert models.Activity.objects.count() == 0
    import_one_activity("cycling_bad_schandau.fit")
    assert models.Activity.objects.count() == 1

    # get activity
    activity = models.Activity.objects.get()
    assert activity.sport.evaluates_for_awards is False
    assert activity.evaluates_for_awards is True
    assert activity_suitable_for_awards(activity) is False

    # best sections got saved
    assert models.BestSection.objects.filter(activity=activity).count() > 0


def test_import_of_activities__evaluates_for_awards(import_one_activity):
    # insert default sport
    sport = models.default_sport()

    assert models.Sport.objects.count() == 1
    assert models.Settings.objects.count() == 1

    # get default sport
    sport = models.Sport.objects.get()
    assert sport.name == "unknown"
    assert sport.evaluates_for_awards is True

    # now import activity and verify that some best sections got saved to the db
    assert models.Activity.objects.count() == 0
    import_one_activity("cycling_bad_schandau.fit")
    assert models.Activity.objects.count() == 1

    # get activity
    activity = models.Activity.objects.get()
    assert activity.sport.evaluates_for_awards is True
    assert activity.evaluates_for_awards is True
    assert activity_suitable_for_awards(activity) is True

    # some best sections got saved
    assert models.BestSection.objects.filter(activity=activity).count() > 0


def test__activity_evaluates_for_awards(insert_activity):
    # in case both activity and sport are suitable, function should also return True
    activity = insert_activity(name="dummy-1", evaluates_for_awards=True)
    assert activity.evaluates_for_awards is True
    assert activity.sport.evaluates_for_awards is True
    assert activity_suitable_for_awards(activity=activity) is True

    # in case only activity is not suitable, function should return False
    activity = insert_activity(name="dummy-2", evaluates_for_awards=False)
    assert activity.evaluates_for_awards is False
    assert activity.sport.evaluates_for_awards is True
    assert activity_suitable_for_awards(activity=activity) is False

    # in case only sport is not suitable, function should return False
    activity = insert_activity(name="dummy", evaluates_for_awards=True)
    assert activity.evaluates_for_awards is True
    sport = activity.sport
    sport.evaluates_for_awards = False
    sport.save()
    assert activity.sport.evaluates_for_awards is False
    assert activity_suitable_for_awards(activity=activity) is False

    # in case both sport and activity are not suitable, function should also return False
    activity = insert_activity(name="dummy", evaluates_for_awards=False)
    assert activity.evaluates_for_awards is False
    sport = activity.sport
    sport.evaluates_for_awards = False
    sport.save()
    assert activity.sport.evaluates_for_awards is False
    assert activity_suitable_for_awards(activity=activity) is False


def test_avoid_unique_constraint_error(tracks_in_tmpdir, transactional_db, caplog):
    settings = models.get_settings()
    copy_demo_fit_files_to_track_dir(
        source_dir=django_settings.INITIAL_TRACE_DATA_DIR,
        targe_dir=settings.path_to_trace_dir,
        list_of_files_to_copy=["cycling_bad_schandau.fit"],
    )

    # create second file with same checksum but different name
    shutil.copy(
        Path(settings.path_to_trace_dir) / "cycling_bad_schandau.fit",
        Path(settings.path_to_trace_dir) / "cycling_bad_schandau2.fit",
    )

    # in rare situations this lead to a unique constraint sql error because
    # of md5sum already being present in db, check that this does not fail
    run_importer__dask(models)

    # check that file importer warns about two files having the same checksum
    assert "The following two files have the same checksum, you might want to remove one of them:" in caplog.text


def test_import_corrupted_fit_file(tracks_in_tmpdir, caplog):
    assert models.Activity.objects.count() == 0
    settings = models.get_settings()

    copy_demo_fit_files_to_track_dir(
        source_dir=django_settings.INITIAL_TRACE_DATA_DIR,
        targe_dir=settings.path_to_trace_dir,
        list_of_files_to_copy=["cycling_bad_schandau.fit"],
    )

    # create a corrupted fit file
    fit = Path(settings.path_to_trace_dir) / "faulty.fit"
    content = "no valid fit file content"
    fit.write_text(content)
    assert fit.read_text() == content

    # importer should not fail
    run_importer__dask(models)

    # one fit file should have been imported
    assert models.Activity.objects.count() == 1

    # but also an error was logged, however the execution should not have failed
    assert "ERROR" in caplog.text
    assert "Failed to parse fit file" in caplog.text


def test_run_importer__single_file(db, demo_data_dir, tmpdir, fit_file):
    assert models.Activity.objects.count() == 0
    settings = models.get_settings()
    settings.path_to_trace_dir = tmpdir
    settings.save()

    # test on empty dir
    run_importer__dask(models)
    assert models.Activity.objects.count() == 0

    # test on dir with one file
    activity_file = Path(demo_data_dir) / fit_file
    shutil.copy2(activity_file, tmpdir)
    run_importer__dask(models)
    assert models.Activity.objects.count() == 1


def test_run_importer__three_files(db, demo_data_dir, tmpdir, fit_file, fit_file_a, fit_file_b):
    assert models.Activity.objects.count() == 0
    settings = models.get_settings()
    settings.path_to_trace_dir = tmpdir
    settings.save()

    # test on dir with one file
    activity_file_1 = Path(demo_data_dir) / fit_file
    shutil.copy2(activity_file_1, tmpdir)
    activity_file_2 = Path(demo_data_dir) / fit_file_a
    shutil.copy2(activity_file_2, tmpdir)
    activity_file_3 = Path(demo_data_dir) / fit_file_b
    shutil.copy2(activity_file_3, tmpdir)
    run_importer__dask(models)
    assert models.Activity.objects.count() == 3


def test_run_importer__warns_about_duplicate_files(db, demo_data_dir, tracks_in_tmpdir, caplog):
    assert models.Activity.objects.count() == 0
    settings = models.get_settings()

    # copy the same file two times into two different files (with same checksum)
    file_a = Path(settings.path_to_trace_dir) / "a.fit"
    file_b = Path(settings.path_to_trace_dir) / "b.fit"
    source_file = Path(demo_data_dir) / "cycling_bad_schandau.fit"
    shutil.copy(source_file, file_a)
    shutil.copy(source_file, file_b)

    assert calc_md5(file_a) == calc_md5(file_b)

    # now run file importer and verify that a proper warning is logged
    run_importer__dask(models)

    assert "WARNING" in caplog.text
    assert (
        f"The following two files have the same checksum, you might want to remove one of them:{file_b} and {file_a}"
        in caplog.text
    )


def test_import_files_with_upper_case_ending(db, demo_data_dir, tracks_in_tmpdir):
    assert models.Activity.objects.count() == 0
    settings = models.get_settings()

    # copy two activity files to have upper case file endings
    fit_file = Path(settings.path_to_trace_dir) / "test_fit.FIT"
    gpx_file = Path(settings.path_to_trace_dir) / "test_gpx.GPX"
    source_fit = Path(demo_data_dir) / "cycling_bad_schandau.fit"
    source_gpx = Path(demo_data_dir) / "cycling_walchensee.gpx"
    shutil.copy(source_fit, fit_file)
    shutil.copy(source_gpx, gpx_file)

    assert fit_file.is_file()
    assert gpx_file.is_file()

    # now run file importer and check that files get imported
    run_importer__dask(models)

    assert models.Activity.objects.count() == 2



import datetime
import json
import os

import pandas as pd
from django.conf import settings
from django.utils.duration import duration_microseconds

from wkz.gis.geo import add_elevation_data_to_coordinates
from wkz.tools.utils import sanitize, timestamp_format

gpx_header = """<?xml version="1.0" encoding="UTF-8"?>
<gpx creator="Fabian Gebhart" version="1.1" xmlns="http://www.topografix.com/GPX/1/1"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd"
xmlns:gpxtpx="http://www.garmin.com/xmlschemas/TrackPointExtension/v1"
xmlns:gpxtrkx="http://www.garmin.com/xmlschemas/TrackStatsExtension/v1"
xmlns:gpxx="http://www.garmin.com/xmlschemas/GpxExtensions/v3">"""


def _gpx_file(time, name, track_points, sport):
    return f"""{gpx_header}
    <metadata>
        <time>{time.strftime(timestamp_format)}</time>
        <link href="https://github.com/fgebhart/workoutizer">
            <text>Workoutizer</text>
        </link>
    </metadata>
    <trk>
        <name>{name}</name>
            <type>{sport}</type>
        <trkseg>
            {track_points}
        </trkseg>
    </trk>
</gpx>
"""


def _track_points(coordinates: list, timestamps: list):
    track_points = ""
    for c, ts in zip(coordinates, timestamps):
        if len(c) > 2:
            point = f"""<trkpt lat="{c[1]}" lon="{c[0]}">
                <time>{ts}</time>
                <ele>{c[2]}</ele>
            </trkpt>
            """
        else:
            point = f"""<trkpt lat="{c[1]}" lon="{c[0]}">
                <time>{ts}</time>
            </trkpt>
            """
        track_points += point
    return track_points


def _build_gpx(time, file_name, coordinates: list, timestamps: list, sport: str):
    return _gpx_file(time=time, name=file_name, track_points=_track_points(coordinates, timestamps), sport=sport)


def _fill_list_of_timestamps(start: datetime.date, duration, length: int):
    list_of_timestamps = []
    duration = datetime.timedelta(microseconds=duration_microseconds(duration))
    one_step_of_time = duration / length
    start = datetime.datetime.combine(start, datetime.time(12, 00))
    for i in range(length):
        interval = start + one_step_of_time * i
        strftime = interval.strftime(timestamp_format)
        list_of_timestamps.append(strftime)
    return list_of_timestamps


def save_activity_to_gpx_file(activity):
    file_name = f"{activity.date.date()}_{sanitize(activity.name)}.gpx"
    path = os.path.join(settings.MEDIA_ROOT, file_name)
    coordinates = list(
        zip(
            list(pd.Series(json.loads(activity.trace_file.longitude_list)).ffill().bfill()),
            list(pd.Series(json.loads(activity.trace_file.latitude_list)).ffill().bfill()),
        )
    )
    if json.loads(activity.trace_file.altitude_list):
        coordinates = add_elevation_data_to_coordinates(
            coordinates=coordinates,
            altitude=list(pd.Series(json.loads(activity.trace_file.altitude_list)).ffill().bfill()),
        )
    file_content = _build_gpx(
        time=activity.date,
        file_name=activity.name,
        coordinates=coordinates,
        timestamps=_fill_list_of_timestamps(start=activity.date, duration=activity.duration, length=len(coordinates)),
        sport=activity.sport.name,
    )
    with open(path, "w+") as f:
        f.write(file_content)

    return path



import pytest
from django.urls import reverse
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from wkz.views import DashboardView


def test_custom_400_page(live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.current_url == live_server.url + reverse("home")

    # show that no alert box is visible
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.CLASS_NAME, "alert-danger")

    # try to open non existing page
    webdriver.get(live_server.url + "/does-not-exist")

    # verify that we got redirected to home
    assert webdriver.current_url == live_server.url + reverse("home")

    # verify that the alert box is visible
    webdriver.find_element(By.CLASS_NAME, "alert-danger")


def test_custom_500_page(live_server, webdriver, monkeypatch):
    # mock the DashboardView.get method to provoke a 500 error, get() should return a rendered html, not a string
    def faulty_get_func(self, request):
        return "dummy-string"

    monkeypatch.setattr(DashboardView, "get", faulty_get_func)

    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.current_url == live_server.url + reverse("home")

    # verify that the alert box is visible
    assert "Error while loading" in webdriver.find_element(By.CLASS_NAME, "alert-danger").text

    # verify the correct title of the card
    assert webdriver.find_element(By.CLASS_NAME, "card-title").text == "Error 🤕"



import django_eventstream
from django.urls import include, path, re_path

from wkz import activity_views, api, awards_views, sport_views, views

urlpatterns = [
    # home Dashboard
    path("", views.DashboardView.as_view(), name="home"),
    path("set_number_of_days/<slug:number_of_days>", views.set_number_of_days, name="set-number-of-days"),
    path("activities_page/<slug:page>", views.get_bulk_of_rows_for_next_page, name="activities-page"),
    # Settings
    path("settings/", views.settings_view, name="settings"),
    path("settings/form", views.settings_form, name="settings-form"),
    path("settings/delete-demo-data/", activity_views.DemoActivityDeleteView.as_view(), name="delete-demo-data"),
    # Help
    path("help/", views.HelpView.as_view(), name="help"),
    # Activities
    path("activity/<slug:activity_id>", activity_views.ActivityView.as_view(), name="activity"),
    path("activity/<slug:activity_id>/edit/", activity_views.edit_activity_view, name="edit-activity"),
    path("activity/<slug:activity_id>/download/", activity_views.download_activity, name="download-activity"),
    path("add-activity/", activity_views.add_activity_view, name="add-activity"),
    re_path(r"^activity/(?P<pk>\d+)/delete/$", activity_views.ActivityDeleteView.as_view(), name="delete-activity"),
    # Sports
    path("sport/<slug:sports_name_slug>", sport_views.SportView.as_view(), name="sport"),
    path("sports/", sport_views.SportsView.as_view(), name="sports"),
    path("add-sport/", sport_views.add_sport_view, name="add-sport"),
    path("sport/<slug:sports_name_slug>/edit/", sport_views.edit_sport_view, name="edit-sport"),
    re_path(r"^sport/(?P<slug>[a-zA-Z0-9-]+)/delete/$", sport_views.SportDeleteView.as_view(), name="delete-sport"),
    # Best Sections
    path("awards/", awards_views.AwardsViews.as_view(), name="awards"),
    # Rest API endpoints
    path("mount-device/", api.mount_device_endpoint),
    path("stop/", api.stop_django_server),
    # events channel
    path("events/", include(django_eventstream.urls), {"channels": ["event"]}),
]



import datetime
import logging
from dataclasses import dataclass

import pandas as pd
import pytz
from django.conf import settings
from fitparse import FitFile

from wkz import configuration
from wkz.file_helper.parser import Parser

log = logging.getLogger(__name__)


class FITParser(Parser):
    def __init__(self, path_to_file: str, md5sum: str):
        self.fit = None
        super(FITParser, self).__init__(path_to_file, md5sum)

    def _parse_metadata(self):
        self.file_name = self.get_file_name_from_path(self.path_to_file)
        self.fit = FitFile(self.path_to_file)

    def _parse_records(self):
        for record in self.fit.get_messages():
            record = record.get_values()
            # print(record)

            # parse laps
            if record.get("event") == "lap":
                lap = _parse_lap_data(record)
                if (lap.end_lat and lap.end_long) or (lap.start_lat and lap.start_long):
                    # only save lap if it comes with at least one coordinate
                    self.laps.append(lap)

            # parse list attributes
            timestamp = record.get("timestamp")
            self.timestamps_list.append(timestamp.timestamp() if timestamp else timestamp)
            self.distance_list.append(record.get("distance"))
            self.longitude_list.append(_to_coordinate(record.get("position_long")))
            self.latitude_list.append(_to_coordinate(record.get("position_lat")))
            altitude = record.get("altitude")
            self.altitude_list.append((float(altitude) / 10) if altitude else altitude)
            self.heart_rate_list.append(record.get("heart_rate"))
            self.temperature_list.append(record.get("temperature"))
            self.cadence_list.append(record.get("cadence"))
            self.speed_list.append(record.get("enhanced_speed"))

            # get first value of records
            if not self.sport:
                self.sport = record.get("sport")
            if not self.aerobic_training_effect:
                self.aerobic_training_effect = record.get("total_training_effect")
            if not self.anaerobic_training_effect:
                self.anaerobic_training_effect = record.get("total_anaerobic_training_effect")

            # get last value of records
            distance = record.get("total_distance")
            if distance:
                self.distance = round(float(distance) / 1000, 2)
            duration = record.get("total_elapsed_time")
            if duration:
                self.duration = datetime.timedelta(seconds=int(duration))
            date = record.get("timestamp")
            if date:
                self.date = date.replace(tzinfo=pytz.timezone(settings.TIME_ZONE))
            calories = record.get("total_calories")
            if calories:
                self.calories = calories
            avg_heart_rate = record.get("avg_heart_rate")
            if avg_heart_rate:
                self.avg_heart_rate = avg_heart_rate
            avg_cadence = record.get("avg_running_cadence")
            if avg_cadence:
                self.avg_cadence = avg_cadence
            avg_speed = record.get("enhanced_avg_speed")
            if avg_speed:
                self.avg_speed = avg_speed
            avg_temperature = record.get("avg_temperature")
            if avg_temperature:
                self.avg_temperature = avg_temperature
            total_ascent = record.get("total_ascent")
            if total_ascent:
                self.total_ascent = total_ascent
            total_descent = record.get("total_descent")
            if total_descent:
                self.total_descent = total_descent

        log.debug(f"found date: {self.date}")
        log.debug(f"found number of coordinates: {len(self.longitude_list)}")
        log.debug(f"found number of altitude: {len(self.altitude_list)}")
        log.debug(f"found number of timestamps: {len(self.timestamps_list)}")
        log.debug(f"found avg_speed: {self.avg_speed}")
        log.debug(f"found sport: {self.sport}")
        log.debug(f"found distance: {self.distance} km")
        log.debug(f"found duration: {self.duration} min")
        log.debug(f"found avg_cadence: {self.avg_cadence} steps/min")
        log.debug(f"found avg_temperature: {self.avg_temperature} Celsius")
        log.debug(f"found number of laps: {len(self.laps)}")
        log.debug(f"found total ascent: {self.total_ascent} and descent: {self.total_descent}")

    def _save_data_to_dataframe(self):
        df = pd.DataFrame(
            {
                "latitude_list": self.latitude_list,
                "longitude_list": self.longitude_list,
                "timestamps_list": self.timestamps_list,
                "distance_list": self.distance_list,
                "altitude_list": self.altitude_list,
                "heart_rate_list": self.heart_rate_list,
                "cadence_list": self.cadence_list,
                "speed_list": self.speed_list,
                "temperature_list": self.temperature_list,
            }
        )
        self.dataframe = df

    def _drop_rows_where_all_records_are_null(self):
        # drop rows where all values are null
        self.dataframe.dropna(how="all", inplace=True)

        # overwrite self attributes with content of columns again
        self.latitude_list = self.dataframe.latitude_list.to_list()
        self.longitude_list = self.dataframe.longitude_list.to_list()
        self.timestamps_list = self.dataframe.timestamps_list.to_list()
        self.distance_list = self.dataframe.distance_list.to_list()
        self.altitude_list = self.dataframe.altitude_list.to_list()
        self.heart_rate_list = self.dataframe.heart_rate_list.to_list()
        self.cadence_list = self.dataframe.cadence_list.to_list()
        self.speed_list = self.dataframe.speed_list.to_list()
        self.temperature_list = self.dataframe.temperature_list.to_list()

    def _convert_list_of_nones_to_empty_list(self):
        for attribute in configuration.time_series_attributes:
            if self.dataframe[attribute].isna().all():
                setattr(self, attribute, [])

    def _set_min_max_values(self):
        for attribute in configuration.min_max_attributes:
            if self.dataframe[attribute].any():
                name = attribute.replace("_list", "")
                setattr(self, f"max_{name}", round(float(self.dataframe[attribute].max()), 2))
                setattr(self, f"min_{name}", round(float(self.dataframe[attribute].min()), 2))

    def _set_avg_values(self):
        for attribute in configuration.avg_attributes:
            name = f'avg_{attribute.replace("_list", "")}'
            # only data is available in df and avg value is not None
            if self.dataframe[attribute].any() and not getattr(self, name, None):
                setattr(self, name, round(float(self.dataframe[attribute].mean()), 2))

    def _post_process_data(self):
        self._save_data_to_dataframe()
        self._drop_rows_where_all_records_are_null()
        self._set_avg_values()
        self._convert_list_of_nones_to_empty_list()
        self._set_min_max_values()


def _parse_lap_data(record):
    lap = LapData(
        start_time=record["start_time"].replace(tzinfo=pytz.timezone(settings.TIME_ZONE)),
        end_time=record["timestamp"].replace(tzinfo=pytz.timezone(settings.TIME_ZONE)),
        elapsed_time=datetime.timedelta(seconds=record["total_elapsed_time"]),
        trigger=record.get("lap_trigger", "unknown"),
        # lap trigger could be 'manual', 'distance' or 'session_end'
        distance=record["total_distance"],
        start_lat=_to_coordinate(record.get("start_position_lat")),
        start_long=_to_coordinate(record.get("start_position_long")),
        end_lat=_to_coordinate(record.get("end_position_lat")),
        end_long=_to_coordinate(record.get("end_position_long")),
    )

    if lap.elapsed_time and lap.distance:
        lap.speed = round(lap.distance / record["total_elapsed_time"], 2)
    return lap


@dataclass
class LapData:
    start_time: datetime.datetime
    end_time: datetime.datetime
    elapsed_time: datetime.timedelta
    trigger: str
    distance: float = None
    start_lat: float = None
    start_long: float = None
    end_long: float = None
    end_lat: float = None
    speed: float = None


def _to_coordinate(point: int):
    if not point:
        return point
    # coordinates conversion parameter, not sure why this was needed, maybe due to swapping lat with lon?
    ccp = 11930464.71111111

    coordinate = float(point) / ccp
    return coordinate



import os

import pytest

from workoutizer import settings as django_settings


@pytest.fixture
def test_data_dir():
    return os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def demo_data_dir():
    return django_settings.INITIAL_TRACE_DATA_DIR


@pytest.fixture(autouse=True)
def keep_tracks_dir_clean():
    """
    Fixture to ensure no test adds activity files to the default tracks dir.
    """
    if os.path.isdir(django_settings.TRACKS_DIR):
        num_before = os.listdir(django_settings.TRACKS_DIR)
        yield
        num_after = os.listdir(django_settings.TRACKS_DIR)
        assert num_before == num_after
    else:
        yield


@pytest.fixture
def fit_file():
    return "2019-09-14-17-22-05.fit"


@pytest.fixture
def fit_file_a():
    return "2019-09-18-16-02-35.fit"


@pytest.fixture
def fit_file_b():
    return "2019-09-25-16-15-53.fit"



import datetime
import operator

import pytest
from django.urls import reverse
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from tests.utils import delayed_assertion
from wkz import models
from workoutizer import settings as django_settings


def test_settings_page__no_demo_activity(live_server, webdriver):
    models.get_settings()
    webdriver.get(live_server.url + reverse("settings"))

    assert webdriver.find_element_by_class_name("navbar-brand").text == "Settings"

    # verify the text of the input field labels
    input_labels = [link.text for link in webdriver.find_elements_by_class_name("col-md-4")]
    assert "Path to Traces Directory " in input_labels
    input_labels.remove("Path to Traces Directory ")
    assert "Path to Garmin Device " in input_labels
    input_labels.remove("Path to Garmin Device ")
    assert "Delete fit Files after Copying  " in input_labels
    input_labels.remove("Delete fit Files after Copying  ")
    # reimport got removed, assert it is not present
    assert "Reimport all Files " not in input_labels
    # verify that the list is empty after remove all given input labels
    assert len(input_labels) == 0

    question_hover = [
        question.get_attribute("data-original-title")
        for question in webdriver.find_elements(By.CLASS_NAME, "fa-question-circle")
    ]
    hover_text = ""
    for text in question_hover:
        hover_text = f"{hover_text} {text}"

    assert "path to a directory which both contains" in hover_text
    assert "the path to a directory in which your" in hover_text
    assert "this option to delete fit files from your" in hover_text

    # verify no demo activity is present
    assert len(models.Activity.objects.filter(is_demo_activity=True)) == 0
    # no delete demo data button present
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.ID, "delete-demo-data")

    # check that "made with love" text is present
    assert (
        f"© {datetime.datetime.now().date().year}, made with by Fabian Gebhart"
        in webdriver.find_element(By.CLASS_NAME, "credits").text
    )


def test_settings_page__demo_activity_present__delete_it(import_demo_data, live_server, webdriver):
    models.get_settings()
    webdriver.get(live_server.url + reverse("settings"))

    assert webdriver.find_element_by_class_name("navbar-brand").text == "Settings"

    # verify no demo activity is present
    assert models.Activity.objects.filter(is_demo_activity=True).count() > 18
    # Demo heading is present
    webdriver.find_element(By.ID, "delete-demo-data")

    # also delete demo activity button is present
    first_delete_button = webdriver.find_element_by_id("delete-demo-data")
    assert first_delete_button.text == "  DELETE"

    # click button to verify demo data gets deleted
    first_delete_button.click()
    assert webdriver.current_url == live_server.url + reverse("delete-demo-data")
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Delete Demo Activities"

    # on the new page find the additional delete button and click it
    second_delete_button = webdriver.find_element_by_class_name("btn-space")
    assert second_delete_button.text == "  DELETE"
    second_delete_button.click()

    # verify that all demo data got deleted
    assert models.Activity.objects.filter(is_demo_activity=True).count() == 0

    # no delete demo data button present
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.ID, "delete-demo-data")


def test_settings_page__edit_and_submit_form(live_server, webdriver):
    # get settings and check that all values are at their default configuration
    settings = models.get_settings()
    assert settings.path_to_trace_dir == django_settings.TRACKS_DIR
    assert settings.path_to_garmin_device == "/run/user/1000/gvfs/"
    assert settings.delete_files_after_import is False
    assert settings.number_of_days == 30

    # go to settings page
    webdriver.get(live_server.url + reverse("settings"))
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Settings"

    # Note, that with using htmx to update the settings form, all fields are submitted once their values got changed

    # modify values by inserting into input fields
    trace_dir_input_field = webdriver.find_element(By.ID, "id_path_to_trace_dir")
    trace_dir_input_field.clear()
    trace_dir_input_field.send_keys("some/dummy/path")
    # basically clicking somewhere else to trigger submitting the change
    webdriver.find_element(By.ID, "navigation").click()

    delayed_assertion(lambda: models.get_settings().path_to_trace_dir, operator.eq, "some/dummy/path")

    garmin_device_input_field = webdriver.find_element(By.ID, "id_path_to_garmin_device")
    garmin_device_input_field.clear()
    garmin_device_input_field.send_keys("garmin/dummy/path")
    webdriver.find_element(By.ID, "navigation").click()
    delayed_assertion(lambda: models.get_settings().path_to_garmin_device, operator.eq, "garmin/dummy/path")

    # got removed, should not be accessible
    with pytest.raises(NoSuchElementException):
        webdriver.find_element_by_css_selector("#id_reimporter_updates_all")
    # verify that the number of days field is not present nor editable
    with pytest.raises(NoSuchElementException):
        webdriver.find_element_by_css_selector("#id_number_of_days")

    # again get settings and check that the values are the once entered above
    settings = models.get_settings()
    assert settings.path_to_trace_dir == "some/dummy/path"
    assert settings.path_to_garmin_device == "garmin/dummy/path"

    # did not get changed, since selenium is not able to click into checkbox (?)
    assert settings.delete_files_after_import is False
    # number of days should not be changed
    assert settings.number_of_days == 30

    # this attribute got removed, so verifying that is does no longer exist
    with pytest.raises(AttributeError):
        assert settings.reimporter_updates_all is True

    # reimport button got removed from settings page
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.ID, "reimport-activities").click()



# Generated by Django 3.1.6 on 2021-03-14 18:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wkz", "0004_settings_path_to_activities_on_device"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="settings",
            name="path_to_activities_on_device",
        ),
    ]



import json
from itertools import combinations
from typing import List, Tuple

import pandas as pd
from bokeh.embed import components
from bokeh.layouts import column, gridplot
from bokeh.models import (
    BoxZoomTool,
    CheckboxButtonGroup,
    CrosshairTool,
    CustomJS,
    HoverTool,
)
from bokeh.models.formatters import DatetimeTickFormatter
from bokeh.plotting import figure
from django.conf import settings

from wkz import configuration as cfg
from wkz import models

plot_matrix = {
    "temperature": {
        "color": "OrangeRed",
        "axis": "°C",
        "title": "Temperature",
    },
    "cadence": {
        "color": "MediumSlateBlue",
        "axis": "revolutions/min",
        "title": "Cadence",
    },
    "speed": {
        "color": "darkred",
        "axis": "km/h",
        "title": "Speed",
    },
    "heart_rate": {
        "color": "DarkOrange",
        "axis": "bpm",
        "title": "Heart Rate",
    },
    "altitude": {
        "color": "Green",
        "second_color": "darkseagreen",
        "axis": "m",
        "title": "Altitude",
    },
}


def plot_time_series(activity: models.Activity) -> Tuple[str, str, int]:
    """
    Plotting function to create the time series plots shown in tha activity page. Depending
    on what data is available this creates the following plots:
    - Altitude
    - Heart Rate
    - Speed
    - Cadence
    - Temperature
    All plots share a connected vertical cross hair tools.

    Parameters
    ----------
    activity : models.Activity
        Activity model containing the required activity for which the plots should be generated

    Returns
    -------
    script, div : tuple(str, str, int)
        the html script and div elements as str used to render the plots in the html templates
        and the third element in the tuple is the number of plots to be rendered
    """

    attributes = activity.trace_file.__dict__
    lap_data = models.Lap.objects.filter(trace=activity.trace_file)
    plots = []
    lap_lines = []

    timestamps = pd.to_datetime(
        pd.Series(json.loads(attributes["timestamps_list"]), dtype=float).iloc[:: cfg.every_nth_value], unit="s"
    )
    x_axis = pd.to_datetime(timestamps).dt.tz_localize("utc").dt.tz_convert(settings.TIME_ZONE)
    x_axis = x_axis - x_axis.min()

    box_zoom_tool = BoxZoomTool(dimensions="width")
    for attribute, values in attributes.items():
        if attribute in cfg.attributes_to_create_time_series_plot_for:
            values = pd.Series(json.loads(values), dtype=float).iloc[:: cfg.every_nth_value]
            if values.any():
                attribute = attribute.replace("_list", "")

                p = figure(
                    x_axis_type="datetime",
                    plot_height=int(settings.PLOT_HEIGHT / 2.5),
                    sizing_mode="stretch_width",
                    y_axis_label=plot_matrix[attribute]["axis"],
                    tools="reset",
                )

                if attribute == "speed":
                    # turn speed values from m/s into km/h to be consistent with other speed values
                    values = values.mul(3.6)
                if attribute == "altitude":
                    p.varea(
                        x=x_axis,
                        y1=values,
                        y2=values.min(),
                        color=plot_matrix[attribute]["second_color"],
                        fill_alpha=0.5,
                    )
                    p.line(
                        x_axis,
                        values,
                        line_width=2,
                        color=plot_matrix[attribute]["color"],
                        legend_label=plot_matrix[attribute]["title"],
                    )
                else:
                    p.line(
                        x_axis,
                        values,
                        line_width=2,
                        color=plot_matrix[attribute]["color"],
                        legend_label=plot_matrix[attribute]["title"],
                    )

                # render vertical lap lines
                lap = _add_laps_to_plot(laps=lap_data, plot=p, y_values=values)
                lap_lines += lap

                # add tools to plot
                x_hover = ("Time", "@x")
                hover = HoverTool(
                    tooltips=[(plot_matrix[attribute]["title"], f"@y {plot_matrix[attribute]['axis']}"), x_hover],
                    mode="vline",
                    toggleable=False,
                )
                p.add_tools(hover)
                p.add_tools(box_zoom_tool)

                p.xgrid.grid_line_color = None
                p.legend.location = "top_left"
                p.legend.label_text_font = "Montserrat"
                p.legend.background_fill_alpha = 0.7
                dtf = DatetimeTickFormatter()
                dtf.minutes = ["%M:%S"]
                p.xaxis.formatter = dtf
                p.xaxis.major_label_overrides = {0: "0:00"}
                plots.append(p)

                values.ffill(inplace=True)
                values.bfill(inplace=True)
                x_axis.ffill(inplace=True)
                x_axis.bfill(inplace=True)

    _link_plot_tools(all_plots=plots, x_values=x_axis)

    layout = gridplot(
        plots, sizing_mode="stretch_width", toolbar_location="right", ncols=1, toolbar_options={"logo": None}
    )

    if lap_data:
        layout = _add_button_to_toggle_laps(lap_lines, layout)

    script, div = components(layout)
    return script, div, len(plots)


def _add_button_to_toggle_laps(lap_lines, layout):
    # include button to toggle rendering of laps
    btn = CheckboxButtonGroup(labels=["Show Laps"], active=[0], width=100)

    js = """
        for (line in laps) {
            laps[line].visible = false;
            if (typeof markerGroup != "undefined") {
                markerGroup.removeFrom(map);
                }
        }
        for (i in cb_obj.active) {
            if (cb_obj.active[i] == 0) {
                for (line in laps) {
                    laps[line].visible = true;
                    if (typeof markerGroup != "undefined") {
                        markerGroup.addTo(map);
                        }
                }
            }
        }
        """
    callback = CustomJS(args={"laps": lap_lines, "checkbox": btn}, code=js)
    btn.js_on_change("active", callback)
    layout = column(layout, btn)
    layout.sizing_mode = "stretch_width"
    return layout


def _add_laps_to_plot(laps: list, plot, y_values: list) -> List:
    lap_lines = []
    x_value = pd.Timedelta(seconds=0)
    for lap in laps:
        x_value += lap.elapsed_time
        if lap.trigger == "manual":
            line = plot.line([x_value, x_value], [y_values.min() - 1, y_values.max() + 1], color="grey")
            lap_lines.append(line)
    return lap_lines


def _link_crosshair_and_render_icon_on_map_on_hover(fig1, fig2, x_values):
    cross1 = CrosshairTool(dimensions="height", toggleable=False)
    fig1.add_tools(cross1)
    fig2.add_tools(cross1)

    # js for rendering the sport icon on leaflet map when hovering over bokeh time series plots
    js_move = """
        if(cb_obj.x >= fig.x_range.start && cb_obj.x <= fig.x_range.end &&
           cb_obj.y >= fig.y_range.start && cb_obj.y <= fig.y_range.end)
        {
            cross.spans.height.computed_location = cb_obj.sx

            // determine closest point in list of x_values in order to render
            // the current position when hovering over time series plots
            var closest = x_values.reduce(function (prev, curr) {
                return (Math.abs(curr - cb_obj.x) < Math.abs(prev - cb_obj.x) ? curr : prev);
            });
            var index = x_values.indexOf(closest);

            // call rendering function, defined in activity_map.html
            render_position(index);
        }
        else
        {
            cross.spans.height.computed_location = null
        }
    """
    js_leave = "cross.spans.height.computed_location = null"
    args = {"cross": cross1, "fig": fig1, "x_values": x_values}
    fig1.js_on_event("mousemove", CustomJS(args=args, code=js_move))
    fig1.js_on_event("mouseleave", CustomJS(args=args, code=js_leave))
    args = {"cross": cross1, "fig": fig2, "x_values": x_values}
    fig2.js_on_event("mousemove", CustomJS(args=args, code=js_move))
    fig2.js_on_event("mouseleave", CustomJS(args=args, code=js_leave))


def _link_plot_tools(all_plots: list, x_values: list):
    if len(all_plots) == 1:
        _link_crosshair_and_render_icon_on_map_on_hover(all_plots[0], all_plots[0], x_values=x_values)
    else:
        for combi in combinations(all_plots, 2):
            _link_crosshair_and_render_icon_on_map_on_hover(combi[0], combi[1], x_values=x_values)
            # add x axis range for linking effect of box zoom tool
            combi[0].x_range = combi[1].x_range



import datetime
import logging

import pandas as pd
import pytz
from bokeh.embed import components
from bokeh.models import BoxZoomTool
from bokeh.plotting import figure
from django.conf import settings
from django.utils import timezone

log = logging.getLogger(__name__)


def _plot_activities(activities, sport_model, number_of_days):
    df = pd.DataFrame(list(activities.values("name", "sport", "date", "duration")))
    sports = sport_model.objects.filter(id__in=df["sport"].unique())
    colors = []
    for sport in sports:
        df[sport.name] = df.loc[df["sport"] == sport.id, "duration"]  # add duration of sports in new column each
        colors.append(sport.color)

    df.drop(columns=["sport", "duration", "name"], inplace=True)
    today = timezone.now().replace(tzinfo=pytz.timezone(settings.TIME_ZONE))
    if number_of_days < 9999:
        start = today - datetime.timedelta(days=number_of_days)
    else:
        start = df["date"].min().replace(tzinfo=pytz.timezone(settings.TIME_ZONE))
    date_range = pd.DataFrame({"date": pd.date_range(start=start, end=today, tz=settings.TIME_ZONE)})
    df = pd.concat([df, date_range], sort=True)

    df["date"] = pd.to_datetime(df["date"], utc=True).dt.date
    df = df.groupby("date", as_index=False).sum()  # group date duplicates to ensure actual stacking of vbars
    df["date_name"] = df["date"].astype(str)
    df.fillna(value=pd.Timedelta(seconds=0), inplace=True)

    p = figure(
        x_axis_type="datetime",
        y_axis_type="datetime",
        plot_height=settings.PLOT_HEIGHT,
        sizing_mode="stretch_width",
        tools="hover,reset",
        tooltips="$name @date_name",
        # tooltips="$name @date_name: @$name",
    )

    sports_list = [sport.name for sport in sports]
    p.vbar_stack(
        sports_list,
        x="date",
        width=datetime.timedelta(days=0.8),
        color=colors,
        muted_color=colors,
        muted_alpha=0.2,
        source=df,
    )

    # render zero hours properly
    p.yaxis.major_label_overrides = {0: "0h"}
    p.toolbar.logo = None

    p.add_tools(BoxZoomTool(dimensions="width"))

    return p


def plot_history(activities, sport_model, number_of_days):
    try:
        script, div = components(
            _plot_activities(activities=activities, sport_model=sport_model, number_of_days=number_of_days)
        )
    except AttributeError and TypeError and ValueError as e:
        log.warning(f"Could not render plot. Check if activity data is correct: {e}", exc_info=True)
        script = ""
        div = "Could not render plot - no activity data found."
    return script, div



from wkz.tools.utils import (
    cut_list_to_have_same_length,
    get_local_ip_address,
    limit_string,
    sanitize,
)


def test_sanitize():
    assert sanitize("Lore Ipsum") == "lore-ipsum"
    assert sanitize("some-words") == "some-words"
    assert sanitize("activity1/3") == "activity1-3"
    assert sanitize(1) == "1"
    assert sanitize(0.123) == "0.123"
    assert sanitize(True) == "true"


def test_cut_list_to_have_same_length():
    # mode: cut beginning
    assert cut_list_to_have_same_length([1, 2, 3, 4, 5], [3, 4, 5]) == ([3, 4, 5], [3, 4, 5])
    assert cut_list_to_have_same_length([3, 4, 5], [1, 2, 3, 4, 5]) == ([3, 4, 5], [3, 4, 5])
    assert cut_list_to_have_same_length([3, 4, 5], [3, 4, 5]) == ([3, 4, 5], [3, 4, 5])
    assert cut_list_to_have_same_length([3], []) == ([], [])
    # mode: fill end of shorter list with same as last item
    assert cut_list_to_have_same_length([1, 2, 3], [1], mode="fill end") == ([1, 2, 3], [1, 1, 1])
    assert cut_list_to_have_same_length([1], [1, 2, 3], mode="fill end") == ([1, 1, 1], [1, 2, 3])
    assert cut_list_to_have_same_length([1], [1, 2, 3], mode="fill end", modify_only_list2=True) == ([1], [1, 2, 3])
    assert cut_list_to_have_same_length(["a"], ["a", "b", "c"], mode="fill end") == (["a", "a", "a"], ["a", "b", "c"])
    assert cut_list_to_have_same_length([3, 4, 5], [3, 4, 5], mode="fill end") == ([3, 4, 5], [3, 4, 5])


def test_limit_string():
    assert limit_string(string="12345", max_length=10) == "12345"
    assert limit_string(string="12345", max_length=2) == "1...5"
    assert limit_string(string="12345", max_length=3) == "1...5"
    assert limit_string(string="12345", max_length=4) == "12...45"
    assert limit_string(string="12345", max_length=5) == "12345"
    assert limit_string(string="some super duper long string", max_length=12) == "some s...string"


def test__get_local_ip_address():
    ip_address = get_local_ip_address()
    assert type(ip_address) == str
    assert len(ip_address) >= 8
    assert "." in ip_address



import datetime

from django.urls import reverse
from selenium.webdriver.common.by import By

from workoutizer.__init__ import __version__


def test_help_page(live_server, webdriver):
    webdriver.get(live_server.url + reverse("help"))
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Help"

    card_title = [a.text for a in webdriver.find_elements_by_class_name("card-title")]

    # check version card
    assert "Version" in card_title
    assert webdriver.find_element(By.CLASS_NAME, "badge").text == __version__

    # check source code card
    hrefs = [a.get_attribute("href") for a in webdriver.find_elements_by_tag_name("a")]
    assert "Source Code" in card_title
    assert "https://github.com/fgebhart/workoutizer" in hrefs

    # check keyboard navigation card
    assert "Keyboard Navigation" in card_title
    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    assert "Go to Dashboard" in table_data
    assert "g + d" in table_data

    assert "Go to all Sports" in table_data
    assert "g + s" in table_data

    assert "Go to Awards" in table_data
    assert "g + a" in table_data

    assert "Go to Sport" in table_data
    assert "g + Number" in table_data

    assert "Go to edit Sport/Activity" in table_data
    assert "g + e" in table_data

    assert "Go to Settings" in table_data
    assert "g + ," in table_data

    assert "Go to Help" in table_data
    assert "g + h" in table_data

    assert "Go to add Activity" in table_data
    assert "g + n" in table_data

    assert "Toggle Sidebar" in table_data
    assert "[" in table_data

    # check that "made with love" text is present
    assert (
        f"© {datetime.datetime.now().date().year}, made with by Fabian Gebhart"
        in webdriver.find_element(By.CLASS_NAME, "credits").text
    )



from wkz.demo import copy_demo_fit_files_to_track_dir


def test_copy_demo_fit_files_to_track_dir__basic(tmp_path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()

    target_dir = tmp_path / "target"
    target_dir.mkdir()

    # create some dummy files
    file_1 = src_dir / "file_1.txt"
    file_1.write_text("foo")
    assert file_1.read_text() == "foo"
    file_2 = src_dir / "file_2.txt"
    file_2.write_text("baa")
    assert file_2.read_text() == "baa"

    # copy all files
    copy_demo_fit_files_to_track_dir(source_dir=str(src_dir), targe_dir=str(target_dir))

    copied_file_1 = target_dir / "file_1.txt"
    copied_file_2 = target_dir / "file_2.txt"

    assert copied_file_1.is_file()
    assert copied_file_2.is_file()

    assert copied_file_1.read_text() == "foo"
    assert copied_file_2.read_text() == "baa"


def test_copy_demo_fit_files_to_track_dir__filter_files(tmp_path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()

    target_dir = tmp_path / "target"
    target_dir.mkdir()

    # create some dummy files
    file_1 = src_dir / "file_1.txt"
    file_1.write_text("foo")
    file_2 = src_dir / "file_2.txt"
    file_2.write_text("baa")

    # copy just file_1
    copy_demo_fit_files_to_track_dir(source_dir=src_dir, targe_dir=target_dir, list_of_files_to_copy=[file_1])

    copied_file_1 = target_dir / "file_1.txt"
    copied_file_2 = target_dir / "file_2.txt"

    assert copied_file_1.is_file()
    assert copied_file_1.read_text() == "foo"
    assert not copied_file_2.is_file()


def test_copy_demo_fit_files_to_track_dir__with_pre_existing_dir(tmp_path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()

    target_dir = tmp_path / "target"
    target_dir.mkdir()

    # add traces dir within target and src dir
    trace_dir = target_dir / "traces"
    trace_dir.mkdir()
    src_trace_dir = src_dir / "traces"
    src_trace_dir.mkdir()

    # create some dummy files
    file_1 = src_trace_dir / "file_1.txt"
    file_1.write_text("foo")
    file_2 = src_trace_dir / "file_2.txt"
    file_2.write_text("baa")

    # copy files (this would fail without 'dirs_exist_ok=True' because trace does exist already)
    copy_demo_fit_files_to_track_dir(source_dir=str(src_dir), targe_dir=str(target_dir))

    copied_file_1 = trace_dir / "file_1.txt"
    copied_file_2 = trace_dir / "file_2.txt"

    assert copied_file_1.is_file()
    assert copied_file_1.read_text() == "foo"
    assert copied_file_2.is_file()
    assert copied_file_2.read_text() == "baa"



lines_colors = [
    "Red",
    "DodgerBlue",
    "LimeGreen",
    "Gold",
    "MediumSlateBlue",
    "Brown",
    "Olive",
    "Orange",
    "DarkGoldenRod",
    "Salmon",
    "Fuchsia",
    "Aqua",
    "LightSlateGray",
    "MediumBlue",
    "GreenYellow",
    "DarkRed",
    "DarkMagenta",
    "Khaki",
    "MediumSpringGreen",
    "OrangeRed",
    "DarkGreen",
    "LightPink",
    "DarkSlateBlue",
    "Yellow",
    "Turquoise",
    "SaddleBrown",
    "Maroon",
    "Ivory",
    "SpringGreen",
    "BlueViolet",
    "Coral",
    "Teal",
    "Navy",
    "LightGoldenRodYellow",
    "DarkOliveGreen",
    "Coral",
    "Red",
    "DodgerBlue",
    "LimeGreen",
    "Gold",
    "MediumSlateBlue",
    "Brown",
    "Olive",
    "Orange",
    "DarkGoldenRod",
    "Salmon",
    "Fuchsia",
    "Aqua",
    "LightSlateGray",
    "MediumBlue",
    "GreenYellow",
    "DarkRed",
    "DarkMagenta",
    "Khaki",
    "MediumSpringGreen",
    "OrangeRed",
    "DarkGreen",
    "LightPink",
    "DarkSlateBlue",
    "Yellow",
    "Turquoise",
    "SaddleBrown",
    "Maroon",
    "Ivory",
    "SpringGreen",
    "BlueViolet",
    "Coral",
    "Teal",
    "Navy",
    "LightGoldenRodYellow",
    "DarkOliveGreen",
    "Coral",
    "Red",
    "DodgerBlue",
    "LimeGreen",
    "Gold",
    "MediumSlateBlue",
    "Brown",
    "Olive",
    "Orange",
    "DarkGoldenRod",
    "Salmon",
    "Fuchsia",
    "Aqua",
    "LightSlateGray",
    "MediumBlue",
    "GreenYellow",
    "DarkRed",
    "DarkMagenta",
    "Khaki",
    "MediumSpringGreen",
    "OrangeRed",
    "DarkGreen",
    "LightPink",
    "DarkSlateBlue",
    "Yellow",
    "Turquoise",
    "SaddleBrown",
    "Maroon",
    "Ivory",
    "SpringGreen",
    "BlueViolet",
    "Coral",
    "Teal",
    "Navy",
    "LightGoldenRodYellow",
    "DarkOliveGreen",
    "Coral",
    "Red",
    "DodgerBlue",
    "LimeGreen",
    "Gold",
    "MediumSlateBlue",
    "Brown",
    "Olive",
    "Orange",
    "DarkGoldenRod",
    "Salmon",
    "Fuchsia",
    "Aqua",
    "LightSlateGray",
    "MediumBlue",
    "GreenYellow",
    "DarkRed",
    "DarkMagenta",
    "Khaki",
    "MediumSpringGreen",
    "OrangeRed",
    "DarkGreen",
    "LightPink",
    "DarkSlateBlue",
    "Yellow",
    "Turquoise",
    "SaddleBrown",
    "Maroon",
    "Ivory",
    "SpringGreen",
    "BlueViolet",
    "Coral",
    "Teal",
    "Navy",
    "LightGoldenRodYellow",
    "DarkOliveGreen",
    "Coral",
    "Red",
    "DodgerBlue",
    "LimeGreen",
    "Gold",
    "MediumSlateBlue",
    "Brown",
    "Olive",
    "Orange",
    "DarkGoldenRod",
    "Salmon",
    "Fuchsia",
    "Aqua",
    "LightSlateGray",
    "MediumBlue",
    "GreenYellow",
    "DarkRed",
    "DarkMagenta",
    "Khaki",
    "MediumSpringGreen",
    "OrangeRed",
    "DarkGreen",
    "LightPink",
    "DarkSlateBlue",
    "Yellow",
    "Turquoise",
    "SaddleBrown",
    "Maroon",
    "Ivory",
    "SpringGreen",
    "BlueViolet",
    "Coral",
    "Teal",
    "Navy",
    "LightGoldenRodYellow",
    "DarkOliveGreen",
    "Coral",
]



import numpy as np
import pandas as pd
from bokeh.embed import components
from bokeh.plotting import figure

from wkz import models


def plot_trend(activities, sport_model):
    number_of_days = models.get_settings().number_of_days

    df = pd.DataFrame.from_records(activities.values("sport_id", "duration", "date"))
    df["date"] = pd.to_datetime(df["date"])
    df = df.set_index("date")
    days = int(number_of_days / 5)
    freq = days if days > 1 else 1
    df = df.groupby([pd.Grouper(freq=f"{freq}D"), "sport_id"]).agg({"duration": np.sum}).reset_index()
    df = df.pivot(index="date", columns="sport_id", values="duration").fillna("0")
    sports = sport_model.objects.exclude(name="unknown").order_by("id").values("id", "name", "color")
    id_color_mapping = {}
    for sport in sports:
        id_color_mapping[sport["id"]] = sport["color"]
    df = df.rename(columns=id_color_mapping)

    p = figure(
        height=135,
        sizing_mode="stretch_width",
        x_axis_type="datetime",
        y_axis_type="datetime",
    )
    p.multi_line(
        xs=[df.index.values] * len(df.columns), ys=[df[name].values for name in df], line_color=df.columns, line_width=3
    )

    # render zero hours properly
    p.yaxis.major_label_overrides = {0: "0h"}
    p.toolbar.logo = None
    p.toolbar_location = None
    p.tools = []

    script_trend, div_trend = components(p)

    return script_trend, div_trend



import datetime

from django import template
from django.db.models.query import QuerySet

register = template.Library()


@register.filter
def duration(td):
    total_seconds = int(td.total_seconds())
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60
    return f"{hours}h {minutes}m"


@register.filter
def hours_only(td):
    total_seconds = int(td.total_seconds())
    hours = total_seconds // 3600
    return f"{hours}h"


@register.filter
def days_to_years(days):
    if int(days) >= 365:
        result = days / 365
    else:
        return f"{days} days"
    if result > 1:
        return f"{int(result)} years"
    else:
        return f"{int(result)} year"


@register.filter
def table_duration(td):
    total_seconds = int(td.total_seconds())
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60
    if minutes < 10:
        minutes = f"0{minutes}"
    return f"{hours}:{minutes}"


@register.filter
def speed_to_pace(speed):
    if speed is None:
        return "00:00"
    if speed <= 0:
        return "00:00"
    pace = (float(speed) * 60 / 1000) ** -1
    delta = datetime.timedelta(minutes=pace)
    hours, remainder = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    pace = "{:02}:{:02}".format(int(minutes), int(seconds))
    return f"{pace}"


@register.filter
def m_per_s_to_km_per_h(m_per_s):
    return round(float(m_per_s) * 3.6, 1)


@register.filter
def round_2nd_decimal(td):
    return round(td, 2)


@register.filter
def to_int(number):
    return int(number)


@register.filter
def h_m_s(delta):
    return str(delta).split(".")[0]


def strfdelta(tdelta, fmt):
    d = {"days": tdelta.days}
    d["hours"], rem = divmod(tdelta.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    return fmt.format(**d)


@register.filter
def queryset_to_list(queryset: QuerySet):
    if isinstance(queryset, QuerySet):
        list_as_string = sorted(queryset.values_list("slug", flat=True))
        if "unknown" in list_as_string:
            list_as_string.remove("unknown")
        return list_as_string


@register.filter
def get_best_sections_list(queryset: QuerySet):
    list_of_start_and_end_index = []
    for section in queryset:
        list_of_start_and_end_index.append([section.start, section.end])
    return list_of_start_and_end_index


@register.filter
def best_section_distance(distance: int) -> str:
    if distance < 1000:
        return f"{distance}m"
    else:
        return f"{int(distance/1000)}km"



from wkz import models


def test_automatic_naming_of_activity__gpx_with_coordinates(import_one_activity):
    import_one_activity("example.gpx")

    activity = models.Activity.objects.get()
    assert activity.name == "Evening Jogging in Heidelberg"


def test_automatic_naming_of_activity__fit_with_coordinates(import_one_activity):
    import_one_activity("hike_with_coordinates_muggenbrunn.fit")

    activity = models.Activity.objects.get()
    assert activity.name == "Noon Hiking in Aftersteg"


def test_automatic_naming_of_activity__fit_no_coordinates(import_one_activity):
    import_one_activity("swim_no_coordinates.fit")

    activity = models.Activity.objects.get()
    assert activity.name == "Noon Swimming"



import os
import subprocess

import pytest
from django.core.management import execute_from_command_line

from wkz import models
from workoutizer import __version__, cli
from workoutizer import settings as django_settings


def test_cli_version():
    output = subprocess.check_output(["wkz", "--version"]).decode("utf-8")
    assert output == f"{__version__}\n"

    output = subprocess.check_output(["wkz", "-v"]).decode("utf-8")
    assert output == f"{__version__}\n"


def test_cli__init(tracks_in_tmpdir):
    cli._init(import_demo_activities=False)
    assert os.path.isdir(django_settings.WORKOUTIZER_DIR)
    assert len(models.Sport.objects.all()) == 0
    assert len(models.Activity.objects.all()) == 0
    assert len(models.Settings.objects.all()) == 1

    cli._init(import_demo_activities=True)
    assert os.path.isdir(django_settings.WORKOUTIZER_DIR)
    assert len(models.Sport.objects.all()) == 5
    assert len(models.Settings.objects.all()) == 1
    assert len(models.Activity.objects.all()) > 1


def test_cli__check__demo_data(tracks_in_tmpdir):
    # initialized wkz with demo data and then run check
    cli._init(import_demo_activities=True)
    cli._check()


def test_cli__check__no_demo_data(db):
    # initialized wkz without demo data and then run check
    cli._init(import_demo_activities=False)
    cli._check()


def test_cli__check__not_initialized(db):
    # if wkz is not initialized expect to raise error - ensure to start with no preexisting db
    execute_from_command_line(["manage.py", "flush", "--noinput"])
    with pytest.raises(cli.NotInitializedError, match="ERROR: Make sure to execute 'wkz init' first"):
        cli._check()


def test_cli__reimport(import_one_activity):
    import_one_activity("cycling_bad_schandau.fit")

    assert models.Activity.objects.count() == 1
    assert models.Settings.objects.count() == 1
    activity = models.Activity.objects.get()
    orig_distance = activity.distance
    activity.distance = 77_777.77
    activity.save()
    assert activity.distance != orig_distance

    cli._reimport()

    activity = models.Activity.objects.get()
    assert activity.distance == orig_distance



import math

import pytest
from sportgems import DistanceTooSmallException, find_fastest_section

from wkz import configuration
from wkz.best_sections.fastest import get_fastest_section


def test_sportgems_fastest_interface__dummy_data():
    # test sportgems interface with the example data given in the repo readme
    coordinates = [(48.123, 9.35), (48.123, 9.36), (48.123, 9.37), (48.123, 9.38)]
    times = [1608228953.8, 1608228954.8, 1608228955.8, 1608228956.8]

    result = find_fastest_section(1000, times, coordinates, tolerance=1000)

    assert result.start == 0
    assert result.end == 3
    assert math.isclose(result.velocity, 495.393, abs_tol=0.01)


def test_sportgems_fastest_interface__real_activity_data__fit(fit_parser):
    parser = fit_parser()
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    result = find_fastest_section(1000, parser.timestamps_list, coordinates)

    assert result.start == 622
    assert result.end == 712
    assert math.isclose(result.velocity, 2.888, abs_tol=0.01)


def test_sportgems_fastest_interface__real_activity_data__gpx(gpx_parser):
    parser = gpx_parser()
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    result = find_fastest_section(1000, parser.timestamps_list, coordinates)

    assert result.start == 57
    assert result.end == 118
    assert math.isclose(result.velocity, 3.051, abs_tol=0.01)


def test_get_fastest_section__fit(fit_parser):
    parser = fit_parser()
    assert parser.distance == 5.84

    # test fastest 1km
    res = get_fastest_section(1000, parser)
    assert res.start == 622
    assert res.end == 712
    assert round(res.max_value, 2) == 2.89

    # test fastest 2km
    res = get_fastest_section(2000, parser)
    assert res.start == 537
    assert res.end == 814
    assert round(res.max_value, 2) == 2.32

    # test fastest 5km
    res = get_fastest_section(5000, parser)
    assert res.start == 76
    assert res.end == 1167
    assert round(res.max_value, 2) == 1.84

    # test fastest 10km, in this case the activity data is shorter
    # than 10km and thus we expect that no suitable section was found
    with pytest.raises(
        DistanceTooSmallException, match="Distance of provided input data is too small for requested desired distance."
    ):
        get_fastest_section(10_000, parser)


def test_get_fastest_section__gpx(gpx_parser):
    parser = gpx_parser()
    assert parser.distance == 4.3

    # test fastest 1km
    res = get_fastest_section(1000, parser)
    assert res.start == 57
    assert res.end == 118
    assert round(res.max_value, 2) == 3.05

    # test fastest 2km
    res = get_fastest_section(2000, parser)
    assert res.start == 54
    assert res.end == 171
    assert round(res.max_value, 2) == 3.04

    # test fastest 5km
    with pytest.raises(
        DistanceTooSmallException, match="Distance of provided input data is too small for requested desired distance."
    ):
        get_fastest_section(5000, parser)


@pytest.mark.parametrize("test_file", ["2020-10-25-10-54-06.fit", "2020-10-25-10-54-06.fit"])
def test_sportgems_fastest_interface__fit_file_which_cause_panic(fit_parser, test_file):
    parser = fit_parser(test_file)
    coordinates = list(zip(parser.latitude_list, parser.longitude_list))
    for distance in configuration.fastest_distances:
        result = find_fastest_section(distance, parser.timestamps_list, coordinates)

        # sanity check that no section causes rust panic
        assert result.end != 0


def test_sportgems_fastest_interface__fit_file_without_coordinate_data(fit_parser):
    parser = fit_parser("swim_no_coordinates.fit")
    # set altitude list to empty list to simulate a fit file without altitude data
    for distance in configuration.fastest_distances:
        result = get_fastest_section(distance, parser)

        # sanity check that no section causes rust panic
        assert result is None



import operator

import pytest
from django.urls import reverse
from selenium.common.exceptions import (
    ElementNotInteractableException,
    NoSuchElementException,
    WebDriverException,
)
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from tests.utils import delayed_assertion
from wkz import models


def test_sidebar(live_server, webdriver):
    def _assert_only_selected_link_is_highlighted(red_link_text: str):
        all_links = webdriver.find_elements(By.TAG_NAME, "a")
        for link in all_links:
            if link.text == red_link_text:
                assert link.value_of_css_property("color") == "rgb(239, 129, 87)"
            else:
                assert link.value_of_css_property("color") != "rgb(239, 129, 87)"

    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Dashboard"
    assert webdriver.current_url == live_server.url + reverse("home")
    _assert_only_selected_link_is_highlighted("DASHBOARD")

    # test element in sidebar
    webdriver.find_element(By.LINK_TEXT, "ADD SPORT").click()
    assert webdriver.current_url == live_server.url + reverse("add-sport")
    _assert_only_selected_link_is_highlighted("ADD SPORT")

    webdriver.find_element(By.LINK_TEXT, "AWARDS").click()
    assert webdriver.current_url == live_server.url + reverse("awards")
    _assert_only_selected_link_is_highlighted("AWARDS")

    # test clicking on "WORKOUTIZER" in the sidebar
    webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER").click()
    assert webdriver.current_url == live_server.url + reverse("home")
    _assert_only_selected_link_is_highlighted("DASHBOARD")

    webdriver.find_element(By.LINK_TEXT, "SPORTS").click()
    assert webdriver.current_url == live_server.url + reverse("sports")
    _assert_only_selected_link_is_highlighted("SPORTS")

    # minimize sidebar
    webdriver.find_element(By.CSS_SELECTOR, ".fa-chevron-left").click()

    # verify dumpbell img is present
    webdriver.find_element(By.TAG_NAME, "img").click()
    assert webdriver.current_url == live_server.url + reverse("home")
    _assert_only_selected_link_is_highlighted("DASHBOARD")

    # now check that both normal logo (text) and mini logo (img) are visible
    webdriver.find_element(By.CLASS_NAME, "logo-normal")
    webdriver.find_element(By.CLASS_NAME, "logo-mini")

    # again minimize sidebar
    webdriver.find_element(By.CSS_SELECTOR, ".fa-chevron-left").click()
    # maximize
    webdriver.find_element(By.CSS_SELECTOR, ".fa-chevron-right").click()
    assert webdriver.current_url == live_server.url + reverse("home")


def test_responsiveness(live_server, webdriver):
    # first go to dashboard page
    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.find_element_by_class_name("navbar-brand").text == "Dashboard"
    assert webdriver.current_url == live_server.url + reverse("home")

    # check home logo is present and interactable
    webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER").click()

    # also check that navbar items are present
    webdriver.find_element(By.ID, "add-activity-button")
    webdriver.find_element(By.ID, "help-button")
    webdriver.find_element(By.ID, "settings-button")

    # see that navbar toggler is not yet interactable
    with pytest.raises(ElementNotInteractableException):
        webdriver.find_element(By.CLASS_NAME, "navbar-toggler").click()
    # same for navbar kebab
    with pytest.raises(ElementNotInteractableException):
        webdriver.find_element(By.CLASS_NAME, "navbar-kebab").click()

    # now change window size to be small in order to trigger the responsiveness of both the sidebar and navbar
    webdriver.set_window_size(800, 600)
    # wait for the sidebar to be collapsed
    WebDriverWait(webdriver, 3).until(EC.invisibility_of_element_located((By.LINK_TEXT, "WORKOUTIZER")))

    # check that home logo has gone
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER")
    # same for other sidebar items
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.LINK_TEXT, "SPORTS")
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.LINK_TEXT, "AWARDS")
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.LINK_TEXT, "ADD SPORT")

    # now toggle the sidebar to expand again
    webdriver.find_element(By.CLASS_NAME, "navbar-toggler").click()
    # sidebar items should be present again
    webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER")
    webdriver.find_element(By.LINK_TEXT, "SPORTS")
    webdriver.find_element(By.LINK_TEXT, "AWARDS")
    webdriver.find_element(By.LINK_TEXT, "ADD SPORT")
    # collapse sidebar again (by clicking into body) and wait until it is gone
    webdriver.find_element(By.ID, "bodyClick").click()
    WebDriverWait(webdriver, 3).until(EC.invisibility_of_element_located((By.ID, "bodyClick")))

    # verify navbar kebab is present and interactable
    webdriver.find_element(By.CLASS_NAME, "navbar-kebab")

    # navbar items are currently not visible
    with pytest.raises(ElementNotInteractableException):
        webdriver.find_element(By.ID, "add-activity-button").click()
    with pytest.raises(ElementNotInteractableException):
        webdriver.find_element(By.ID, "help-button").click()
    with pytest.raises(ElementNotInteractableException):
        webdriver.find_element(By.ID, "settings-button").click()

    # click navbar kebab to expand navbar items
    webdriver.find_element(By.CLASS_NAME, "navbar-kebab").click()

    # verify navbar items are visible and interactable again
    webdriver.find_element(By.ID, "add-activity-button").click()
    webdriver.find_element(By.CLASS_NAME, "navbar-kebab").click()
    webdriver.find_element(By.ID, "help-button").click()
    webdriver.find_element(By.CLASS_NAME, "navbar-kebab").click()
    webdriver.find_element(By.ID, "settings-button").click()


def test_custom_navbar_items(db, live_server, webdriver, import_one_activity):
    default_slugs = ["add-activity", "settings", "help", "awards", "sports", "add-sport"]
    all_possible_slugs = set(default_slugs + ["edit", "download"])
    import_one_activity("cycling_bad_schandau.fit")

    def _assert_that_only_these_slugs_are_present(slugs_to_check: list):
        links = [cell.get_attribute("href") for cell in webdriver.find_elements(By.TAG_NAME, "a")]
        present_slugs = []
        for link in links:
            if link:
                link = link.split("/")[
                    -2
                ]  # get second last entry since url looks e.g. 'http://localhost:45063/add-activity/'
                present_slugs.append(link)
        for slug in slugs_to_check:
            assert slug in present_slugs
        # check that the remaining ones are not present
        for slug in all_possible_slugs - set(slugs_to_check):
            assert slug not in present_slugs

    # check that on dashboard page only add-activity, help and settings are present
    webdriver.get(live_server.url + reverse("home"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for awards page
    webdriver.get(live_server.url + reverse("awards"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for sports page
    webdriver.get(live_server.url + reverse("sports"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for add-activity page
    webdriver.get(live_server.url + reverse("sports"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for help page
    webdriver.get(live_server.url + reverse("help"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for settings page
    webdriver.get(live_server.url + reverse("settings"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # same for add-sport page
    webdriver.get(live_server.url + reverse("add-sport"))
    _assert_that_only_these_slugs_are_present(default_slugs)

    # sport page should have default slugs + edit slug
    webdriver.get(live_server.url + "/sport/unknown")
    _assert_that_only_these_slugs_are_present(default_slugs + ["edit"])

    # activity page should have default slugs + edit + download
    delayed_assertion(models.Activity.objects.count, operator.eq, 1)
    pk = models.Activity.objects.get().pk
    webdriver.get(live_server.url + f"/activity/{pk}")
    _assert_that_only_these_slugs_are_present(default_slugs + ["edit", "download"])


def _try_to_perform_chain(action_chain: ActionChains):
    # sometimes fails with: Component returned failure code: 0x80004005 (NS_ERROR_FAILURE)
    # maybe due to too lager page_source, TODO try removing try/except once js/css got cleaned up
    # https://stackoverflow.com/questions/57598293/webdriverexception-message-ns-error-failure-location-js-frame-chrome
    try:
        action_chain.perform()
    except WebDriverException as e:
        print(f"failed due to : {e}")


@pytest.mark.parametrize(
    "buttons, page",
    [
        (["g", "d"], "home"),
        (["g", "s"], "sports"),
        (["g", "a"], "awards"),
        (["g", ","], "settings"),
        (["g", "h"], "help"),
        (["g", "n"], "add-activity"),
    ],
)
def test_keyboard_shortcuts__go_to(live_server, webdriver, buttons, page):
    webdriver.get(live_server.url + reverse("add-sport"))
    assert webdriver.current_url == live_server.url + reverse("add-sport")
    element = webdriver.find_element(By.CLASS_NAME, "navbar-brand")

    ac = ActionChains(webdriver)
    ac.key_down(buttons[0])
    ac.click(element)
    ac.key_down(buttons[1])
    ac.click(element)
    ac.key_up(buttons[0])
    ac.key_up(buttons[1])

    _try_to_perform_chain(ac)

    assert webdriver.current_url == live_server.url + reverse(page)


def test_keyboard_shortcuts__activity_and_sport(flush_db, live_server, webdriver, insert_sport, insert_activity):
    insert_sport("Bowling")
    insert_activity()

    webdriver.get(live_server.url + "/activity/1")
    assert webdriver.current_url == live_server.url + "/activity/1"
    element = webdriver.find_element(By.CLASS_NAME, "navbar-brand")

    # test going to edit activity
    ac = ActionChains(webdriver)
    ac.key_down("g")
    ac.click(element)
    ac.key_down("e")
    ac.click(element)
    ac.key_up("g")
    ac.key_up("e")
    _try_to_perform_chain(ac)

    assert webdriver.current_url == live_server.url + "/activity/1/edit/"

    webdriver.get(live_server.url + "/sport/bowling")
    assert webdriver.current_url == live_server.url + "/sport/bowling"
    element = webdriver.find_element(By.CLASS_NAME, "navbar-brand")

    # test going to edit sport
    ac = ActionChains(webdriver)
    ac.key_down("g")
    ac.click(element)
    ac.key_down("e")
    ac.click(element)
    ac.key_up("g")
    ac.key_up("e")
    _try_to_perform_chain(ac)

    assert webdriver.current_url == live_server.url + "/sport/bowling/edit/"


def test_keyboard_shortcuts__toggle_sidebar(live_server, webdriver):
    webdriver.get(live_server.url + reverse("home"))
    assert webdriver.current_url == live_server.url + reverse("home")
    element = webdriver.find_element(By.CLASS_NAME, "navbar-brand")

    # check that WORKOUTIZER is visible
    webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER")

    # collapse sidebar
    ac = ActionChains(webdriver)
    ac.key_down("[")
    ac.click(element)
    ac.key_up("[")
    _try_to_perform_chain(ac)

    WebDriverWait(webdriver, 3).until(EC.invisibility_of_element_located((By.LINK_TEXT, "WORKOUTIZER")))

    # check that WORKOUTIZER is now invisible
    with pytest.raises(NoSuchElementException):
        webdriver.find_element(By.LINK_TEXT, "WORKOUTIZER")

    element = webdriver.find_element(By.CLASS_NAME, "navbar-brand")
    # expand sidebar
    ac = ActionChains(webdriver)
    ac.key_down("[")
    ac.click(element)
    ac.key_up("[")
    _try_to_perform_chain(ac)

    # wait and see if WORKOUTIZER will be coming back again
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.LINK_TEXT, "WORKOUTIZER")))



from dataclasses import dataclass


@dataclass
class GenericBestSection:
    distance: int
    start: int
    end: int
    max_value: float
    kind: str


def activity_suitable_for_awards(activity) -> bool:
    if activity.evaluates_for_awards is False or activity.sport.evaluates_for_awards is False:
        return False
    else:
        return True



import logging
import os
from pathlib import Path

from wkz import models
from wkz.file_helper.fit_collector import FitCollector
from wkz.file_importer import run_importer__dask

log = logging.getLogger(__name__)


def trigger_file_watchdog():
    settings = models.get_settings()
    if Path(settings.path_to_trace_dir).is_dir():
        run_importer__dask(models)
    else:
        log.warning(f"File Watchdog: {settings.path_to_trace_dir} is not a valid directory.")


def trigger_device_watchdog():
    settings = models.get_settings()
    log.debug(f"checking for mounted device at '{settings.path_to_garmin_device}' ...")

    _watch_for_device(
        path_to_trace_dir=settings.path_to_trace_dir,
        path_to_garmin_device=settings.path_to_garmin_device,
        delete_files_after_import=settings.delete_files_after_import,
    )


def _watch_for_device(path_to_garmin_device: str, path_to_trace_dir: str, delete_files_after_import: bool):
    device_mounted = False
    if Path(path_to_garmin_device).is_dir():
        sub_dirs = os.listdir(path_to_garmin_device)
        if len(sub_dirs) == 1 and not device_mounted:
            device_mounted = True
            log.info(f"Found mounted device at {path_to_garmin_device}, triggering fit collector...")
            fit_collector = FitCollector(
                path_to_garmin_device=path_to_garmin_device,
                target_location=path_to_trace_dir,
                delete_files_after_import=delete_files_after_import,
            )
            fit_collector.copy_fit_files()
        elif len(sub_dirs) == 0:
            device_mounted = False
    else:
        log.warning(f"Device Watchdog: {path_to_garmin_device} is not a valid directory.")



import datetime
import hashlib
import logging
import socket


log = logging.getLogger(__name__)

timestamp_format = "%Y-%m-%dT%H:%M:%SZ"


def sanitize(text):
    return str(text).lower().replace(" ", "-").replace("/", "-")


def calc_md5(file) -> str:
    hash_md5 = hashlib.md5()
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def cut_list_to_have_same_length(list1, list2, mode="cut beginning", modify_only_list2=False):
    diff = len(list1) - len(list2)
    if mode == "cut beginning":
        if diff < 0:
            list2 = list2[abs(diff) :]
        elif diff > 0:
            if not modify_only_list2:
                list1 = list1[diff:]
    elif mode == "fill end":
        if diff < 0:  # last 2 is larger
            if not modify_only_list2:
                list1 = list1 + abs(diff) * [list1[-1]]
        elif diff > 0:  # list 1 is larger
            list2 = list2 + diff * [list2[-1]]
    else:
        raise NotImplementedError("mode not implemented")
    return list1, list2


def remove_microseconds(delta):
    return delta - datetime.timedelta(microseconds=delta.microseconds)


def limit_string(string: str, max_length: int):
    string = str(string)
    if max_length >= len(string):
        return string
    else:
        return f"{string[:(int(max_length/2))]}...{string[(-int(max_length/2)):]}"


def get_local_ip_address() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip_address = s.getsockname()[0]
    s.close()
    return ip_address



from sportgems import find_fastest_section

from wkz.best_sections.generic import GenericBestSection


def get_fastest_section(distance: int, parser) -> GenericBestSection:
    # safety check to catch activities with missing coordinate data
    if not parser.latitude_list or not parser.longitude_list:
        return None

    coordinates = list(zip(parser.latitude_list, parser.longitude_list))

    # call to rust
    section = find_fastest_section(distance, parser.timestamps_list, coordinates)

    return GenericBestSection(
        start=section.start,
        end=section.end,
        distance=distance,
        max_value=round(section.velocity, 2),
        kind="fastest",
    )



import datetime

import pytz
from django.urls import reverse
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from wkz import configuration, models


def test_all_sports_page_accessible(live_server, webdriver):
    webdriver.get(live_server.url + reverse("sports"))

    # first time running workoutizer will lead to the dashboard page with no data
    card_title = webdriver.find_element_by_class_name("card-title")
    assert card_title.text == "Sports Overview"
    assert "Sports" in webdriver.find_element_by_class_name("navbar-brand").text


def test_adding_new_sport(live_server, webdriver):
    models.default_sport()
    # verify that the only available sport is the default 'unknown' sport
    sports = models.Sport.objects.all()
    assert len(sports) == 1
    assert sports[0].name == "unknown"
    # webdriver.get(live_server.url + reverse("sports"))
    webdriver.get(live_server.url + reverse("add-sport"))

    # fill out form to add new sport
    sport_name_input_field = webdriver.find_element_by_css_selector("#id_name")
    sport_name_input_field.clear()
    sport_name_input_field.send_keys("UltimateFrisbee")

    icon_input_field = webdriver.find_element_by_css_selector("#id_icon")
    icon_input_field.clear()
    icon_input_field.send_keys("compact-disc")

    # TODO: color seems not to work that way, will leave it for now...
    color_input_field = webdriver.find_element_by_css_selector("#id_color")
    color_input_field.clear()
    color_input_field.send_keys("#FF1F3D")

    # find button and submit
    button = webdriver.find_element_by_id("button")
    button.click()

    # check that a new sport was added
    sports = models.Sport.objects.all()
    assert len(sports) == 2

    new_sport = models.Sport.objects.get(name="UltimateFrisbee")
    assert new_sport.icon == "compact-disc"
    # TODO check why color can not be changed that way
    # assert new_sport.color == "#FF1F3D"


def test_sport_page__complete(import_demo_data, live_server, webdriver):
    # import demo activity and check that all expected elements are available on the sport page
    webdriver.get(live_server.url + "/sport/cycling")

    table_header = [cell.text for cell in webdriver.find_elements_by_tag_name("th")]
    assert "DATE" in table_header
    assert "ACTIVITY" in table_header
    assert "TRACK" in table_header
    assert "SPORT" in table_header
    assert "DURATION" in table_header
    assert "DISTANCE" in table_header

    # the sport name should be unknown
    assert "Cycling" in webdriver.find_element_by_class_name("navbar-brand").text

    card_category = [a.text for a in webdriver.find_elements_by_class_name("card-category")]
    assert "Trend" in card_category
    assert "Distance" in card_category
    assert "Duration" in card_category
    assert "Count" in card_category

    card_title = [a.text for a in webdriver.find_elements_by_class_name("card-title")]
    # assert "0h 0m" in card_title        # fails in CI
    assert "169 km" in card_title
    assert "13h" in card_title
    assert "3" in card_title

    links = [a.text for a in webdriver.find_elements_by_tag_name("a")]
    assert "CYCLING" in links

    card_title = [p.text for p in webdriver.find_elements_by_class_name("card-title")]
    assert "Overview" in card_title

    paragraph = [p.text for p in webdriver.find_elements_by_tag_name("p")]
    assert "DASHBOARD" in paragraph
    assert "AWARDS" in paragraph
    assert "SPORTS" in paragraph
    assert "CYCLING" in paragraph
    assert "HIKING" in paragraph
    assert "JOGGING" in paragraph
    assert "ADD SPORT" in paragraph

    # wait until loading image is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "loading-bar")))
    # wait until activity row is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "activities-table-row")))

    table_data = [cell.text for cell in webdriver.find_elements_by_tag_name("td")]
    assert "Noon Cycling in Bad Schandau" in table_data
    assert "Early Morning Cycling in Kochel am See" in table_data
    assert "Noon Cycling in Hinterzarten" in table_data

    assert len(webdriver.find_elements_by_class_name("fa-chart-line")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-trophy")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-road")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-history")) > 0
    assert len(webdriver.find_elements_by_class_name("fa-hashtag")) > 0

    # check that map is displayed
    map_text = webdriver.find_element_by_id("leaflet_map").text
    assert "Streets" in map_text
    assert "Leaflet | Map data: © OpenStreetMap" in map_text
    assert "−" in map_text
    assert "+" in map_text
    assert "Satellite" in map_text
    assert "Terrain" in map_text
    assert "Topo" in map_text

    # check that bokeh plot is available
    assert len(webdriver.find_elements_by_class_name("bk-canvas")) == 1

    # check that it is possible to click on the fullscreen toggle using leaflet-ui
    webdriver.find_element_by_css_selector(".leaflet-control-zoom-fullscreen").click()


def test_sport_page__infinite_scroll(live_server, webdriver, insert_activity, insert_sport):
    rows_per_page = configuration.number_of_rows_per_page_in_table
    # insert more activities than the currently configured value of rows
    # per page in order to be  able to trigger the htmx ajax request
    sport = insert_sport(name="Skiing")
    nr_of_inserted_activities = rows_per_page + 5
    for i in range(nr_of_inserted_activities):
        insert_activity(name=f"Dummy Activity {i}", sport=sport)

    assert models.Activity.objects.filter(sport__name="Skiing").count() == nr_of_inserted_activities
    webdriver.get(live_server.url + "/sport/skiing")

    # number of rows equals the number of rows per page, since only one page is loaded
    table_rows = [cell.text for cell in webdriver.find_elements_by_id("activities-table-row")]
    assert len(table_rows) + 1 == rows_per_page

    webdriver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
    # wait until loading image is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "loading-bar")))
    # wait until final row indicating that no more activities are available is present
    WebDriverWait(webdriver, 3).until(EC.presence_of_element_located((By.ID, "end-of-activities")))

    # again check number of table rows
    table_rows = [cell.text for cell in webdriver.find_elements_by_id("activities-table-row")]
    assert len(table_rows) + 1 == nr_of_inserted_activities


def test_sport_page__no_activities_selected_for_plot(live_server, webdriver, insert_activity, insert_sport):
    sport = insert_sport(name="Bungee Jumping")
    for i in range(5):
        date = datetime.datetime(1900, 1, 1, tzinfo=pytz.UTC) + datetime.timedelta(days=i)
        insert_activity(name=f"Dummy Activity {i}", sport=sport, date=date)

    assert models.Activity.objects.filter(sport__slug="bungee-jumping").count() == 5
    webdriver.get(live_server.url + "/sport/bungee-jumping")

    # even though the activity data is far in the past, the rows will be shown in the table
    table_rows = [cell.text for cell in webdriver.find_elements_by_id("activities-table-row")]
    assert len(table_rows) == 5

    paragraph = [p.text for p in webdriver.find_elements_by_tag_name("p")]
    assert "BUNGEE JUMPING" in paragraph
    # because the activity was added has dates far in the past, there is no activity data available for plotting
    assert (
        "Either increase the selected time range, or do some sports and add it to Workoutizer.\n"
        "The activity plot will appear here  "
    ) in paragraph

    links = [a.text for a in webdriver.find_elements_by_tag_name("a")]
    assert "WORKOUTIZER" in links

    # verify leaflet elements are present
    assert "+" in links
    assert "−" in links
    assert "Leaflet" in links
    assert "OpenStreetMap" in links

    spans = [a.text for a in webdriver.find_elements_by_tag_name("span")]
    assert "Streets" in spans
    assert "Topo" in spans
    assert "Terrain" in spans
    assert "Satellite" in spans



import datetime

import pytest
import pytz

from wkz import configuration, models
from wkz.best_sections.generic import activity_suitable_for_awards
from wkz.demo import prepare_import_of_demo_activities
from wkz.file_importer import run_importer__dask


def test_reimport_of_activities(tracks_in_tmpdir, client):
    """
    Test reimporter in following steps:
    1. import demo activities
    2. modify some attributes of a given activity
    3. trigger reimporter
    4. check that attributes have been overwritten with the original values
    5. check that activity page is accessible
    """

    # 1. import one cycling and one hiking activities
    prepare_import_of_demo_activities(
        models,
        list_of_files_to_copy=[
            "hike_with_coordinates_muggenbrunn.fit",
            "cycling_bad_schandau.fit",
        ],
    )
    assert len(models.Sport.objects.all()) == 5
    assert len(models.Settings.objects.all()) == 1

    run_importer__dask(models, importing_demo_data=True)
    all_activities = models.Activity.objects.all()
    assert len(all_activities) == 11
    assert len(models.Activity.objects.filter(sport__slug="swimming")) == 9
    assert len(models.Activity.objects.filter(sport__slug="jogging")) == 0

    all_cycling = models.Activity.objects.filter(sport__slug="cycling")
    assert len(all_cycling) == 1
    cycling = all_cycling[0]
    orig_cycling_distance = cycling.distance
    orig_cycling_duration = cycling.duration
    orig_cycling_name = cycling.name
    orig_cycling_date = cycling.date

    cycling_best_sections = models.BestSection.objects.filter(activity=cycling.pk)
    orig_number_of_cycling_best_sections = len(cycling_best_sections)

    all_hiking = models.Activity.objects.filter(sport__slug="hiking")
    assert len(all_hiking) == 1
    hiking = all_hiking[0]
    orig_hiking_distance = hiking.distance
    orig_hiking_duration = hiking.duration
    orig_hiking_name = hiking.name

    # check that min and max altitude got imported, related to bug fix
    assert hiking.trace_file.max_altitude is not None
    assert hiking.trace_file.min_altitude is not None

    hiking_fastest_sections = models.BestSection.objects.get(activity=hiking.pk, distance=1000, kind="fastest")
    orig_1km_fastest_start = hiking_fastest_sections.start
    orig_1km_velocity = hiking_fastest_sections.max_value

    hiking_climb_sections = models.BestSection.objects.get(activity=hiking.pk, distance=200, kind="climb")
    orig_1km_climb_start = hiking_climb_sections.start
    orig_1km_climb = hiking_climb_sections.max_value

    # 2. modify some attributes of a given activity
    new_date = datetime.datetime(1999, 1, 1, 19, 19, 19, tzinfo=pytz.utc)

    hiking.distance = 5_000.0
    hiking.duration = datetime.timedelta(hours=500)
    hiking.name = "some arbitrary hiking name"
    # remove the demo activity flag of hiking activity
    hiking.is_demo_activity = False
    hiking.date = new_date
    hiking.save()

    hiking_fastest_sections.start = 50_000_000
    hiking_fastest_sections.max_value = 999.999
    hiking_fastest_sections.save()

    hiking_climb_sections.start = 70_000_000
    hiking_climb_sections.max_value = 12345.6789
    hiking_climb_sections.save()

    # verify values got changed
    hiking_climb_sections_modified = models.BestSection.objects.get(activity=hiking.pk, distance=200, kind="climb")
    assert hiking_climb_sections_modified.start == 70_000_000
    assert hiking_climb_sections_modified.max_value == 12345.6789

    cycling.distance = 9_000.0
    cycling.duration = datetime.timedelta(hours=900)
    cycling.name = "some arbitrary cycling name"
    cycling.date = new_date
    cycling.save()

    # verify that cycling is a demo activity
    assert cycling.is_demo_activity is True

    # get lap values to verify reimporter handles updating of lap data correctly
    lap_data = models.Lap.objects.filter(trace=cycling.trace_file)
    orig_lap_speeds = [lap.speed for lap in lap_data]
    # modify lap speed
    for lap in lap_data:
        lap.speed = 123456.789
        lap.save()

    # verify it got changed
    assert [lap.speed for lap in models.Lap.objects.filter(trace=cycling.trace_file)] != orig_lap_speeds

    # delete all cycling best sections
    for best_section in cycling_best_sections:
        best_section.delete()

    assert len(models.BestSection.objects.filter(activity=cycling.pk)) == 0

    # 3. trigger reimport to update values
    run_importer__dask(models, reimporting=True)

    all_activities = models.Activity.objects.all()
    assert len(all_activities) == 11

    # 4. check that attributes have been overwritten with the original values
    updated_hiking = models.Activity.objects.get(sport__slug="hiking")
    assert updated_hiking.distance == orig_hiking_distance
    assert updated_hiking.duration == orig_hiking_duration
    # names should not be overwritten
    assert updated_hiking.name != orig_hiking_name
    assert updated_hiking.name == "some arbitrary hiking name"
    # the date should be not stay at its new value because the hiking activity is not a demo activity it should rather
    # be updated back to its original value, but thats hardly possible since the date of demo activities is adjusted to
    # reflect the current time
    assert updated_hiking.date != new_date

    # verify that attributes of best section got overwritten
    updated_hiking_fastest_sections = models.BestSection.objects.get(
        activity=updated_hiking.pk, distance=1000, kind="fastest"
    )
    assert updated_hiking_fastest_sections.start == orig_1km_fastest_start
    assert updated_hiking_fastest_sections.max_value == orig_1km_velocity

    updated_hiking_climb_sections = models.BestSection.objects.get(
        activity=updated_hiking.pk, distance=200, kind="climb"
    )
    assert updated_hiking_climb_sections.start == orig_1km_climb_start
    assert updated_hiking_climb_sections.max_value == orig_1km_climb

    updated_cycling = models.Activity.objects.get(sport__slug="cycling")
    assert updated_cycling.distance == orig_cycling_distance
    assert updated_cycling.duration == orig_cycling_duration
    # names should not be overwritten
    assert updated_cycling.name != orig_cycling_name
    assert updated_cycling.name == "some arbitrary cycling name"
    # the date of an demo activity should also not updated back to its original
    assert updated_cycling.date != orig_cycling_date
    assert updated_cycling.date == new_date
    # verify that cycling is still a demo activity
    assert updated_cycling.is_demo_activity is True

    # verify that all cycling best sections got reimported and created again
    assert len(models.BestSection.objects.filter(activity=cycling.pk)) == orig_number_of_cycling_best_sections

    # verify lap data is back to original speed values
    updated_lap_data = models.Lap.objects.filter(trace=cycling.trace_file)
    updated_lap_speeds = [lap.speed for lap in updated_lap_data]
    assert updated_lap_speeds == orig_lap_speeds

    # 5. verify that the activity pages are accessible after reimporting
    activities = all_activities
    for activity in activities:
        response = client.get(f"/activity/{activity.pk}")
        assert response.status_code == 200


@pytest.mark.parametrize("kind", ["fastest", "climb"])
def test_reimporting_of_best_sections(import_one_activity, kind):
    # import one cycling activity
    import_one_activity("cycling_bad_schandau.fit")

    assert models.Activity.objects.count() == 1
    assert models.Settings.objects.count() == 1

    run_importer__dask(models)
    assert models.Activity.objects.count() == 1

    activity = models.Activity.objects.get()
    bs = models.BestSection.objects.filter(activity=activity, kind=kind)

    # there should never be more best sections of kind 'fastest' than configured possible fastest sections
    assert len(bs) <= len(configuration.fastest_distances)

    # store original values
    orig_start = [section.start for section in bs]
    orig_end = [section.end for section in bs]
    orig_max_values = [section.max_value for section in bs]
    orig_number_of_best_sections = len(bs)

    # modify values
    for section in bs:
        section.start = 10_000
        section.end = 20_000
        section.max_value = 33_333.3
        section.save()

    # verify that the data got changed
    changed_bs = models.BestSection.objects.filter(activity=activity, kind=kind)
    assert [section.start for section in changed_bs] != orig_start
    assert [section.end for section in changed_bs] != orig_end
    assert [section.max_value for section in changed_bs] != orig_max_values

    # also add another dummy best section which should be removed again by the reimport
    dummy_section = models.BestSection(
        activity=activity,
        kind=kind,
        distance=12345,
        start=42,
        end=84,
        max_value=999.999,
    )
    dummy_section.save()

    # verify number of sections has increased
    assert len(models.BestSection.objects.filter(activity=activity, kind=kind)) == orig_number_of_best_sections + 1

    # now trigger reimport to update modified values
    run_importer__dask(models, reimporting=True)

    # check that dummy section was deleted because it is not present in the configured fastest sections
    assert len(models.BestSection.objects.filter(activity=activity, kind=kind)) == orig_number_of_best_sections

    # verify that the modified values are back to their original values
    # verify that the data got changed
    updated_bs = models.BestSection.objects.filter(activity=activity, kind=kind)
    assert [section.start for section in updated_bs] == orig_start
    assert [section.end for section in updated_bs] == orig_end
    assert [section.max_value for section in updated_bs] == orig_max_values

    if kind == "fastest":
        for section in updated_bs:
            assert section.distance in configuration.fastest_distances
    elif kind == "climb":
        for section in updated_bs:
            assert section.distance in configuration.climb_distances


def test_reimport__not_evaluates_for_awards__changing_sport_flag(import_one_activity):
    # Changed behaviour of this test to check that best sections do not get removed when changing evaluates_for_awards

    import_one_activity("cycling_bad_schandau.fit")

    # verify activity is suitable for best sections
    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is True
    assert models.BestSection.objects.filter(activity=activity).count() > 0

    # change sport flag for evaluates_for_awards to False
    sport = activity.sport
    sport.evaluates_for_awards = False
    sport.save()
    assert activity_suitable_for_awards(activity) is False

    # reimport activity
    run_importer__dask(models, reimporting=True)

    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is False

    # check that best sections did not get removed
    assert models.BestSection.objects.filter(activity=activity).count() > 0

    # now change everything back and verify that the best sections get saved to db again by reimporting
    sport = activity.sport
    sport.evaluates_for_awards = True
    sport.save()
    assert activity_suitable_for_awards(activity) is True

    # reimport activity
    run_importer__dask(models, reimporting=True)

    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is True

    # check that best sections got removed
    assert models.BestSection.objects.filter(activity=activity).count() > 0


def test_reimport__not_evaluates_for_awards__changing_activity_flag(import_one_activity):
    # Changed behaviour of this test to check that best sections do not get removed when changing evaluates_for_awards

    import_one_activity("cycling_bad_schandau.fit")

    # verify activity is suitable for best sections
    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is True
    assert models.BestSection.objects.filter(activity=activity).count() > 0

    # change activity flag for evaluates_for_awards to False
    activity.evaluates_for_awards = False
    activity.save()
    assert activity_suitable_for_awards(activity) is False

    # reimport activity
    run_importer__dask(models, reimporting=True)

    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is False

    # check that best sections did not get removed
    assert models.BestSection.objects.filter(activity=activity).count() > 0

    # now change everything back and verify that the best sections get saved to db again by reimporting
    activity.evaluates_for_awards = True
    activity.save()
    assert activity_suitable_for_awards(activity) is True

    # reimport activity
    run_importer__dask(models, reimporting=True)

    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    assert activity_suitable_for_awards(activity) is True

    # check that best sections got removed
    assert models.BestSection.objects.filter(activity=activity).count() > 0


def test_run_importer__dask__reimporting(tmp_path, import_one_activity):
    import_one_activity("cycling_bad_schandau.fit")
    assert models.Activity.objects.count() == 1
    activity = models.Activity.objects.get()
    original_distance = activity.distance

    # change some values
    activity.name = "Foo"
    activity.distance = 999.9
    activity.save()

    activity = models.Activity.objects.get()
    assert activity.name == "Foo"

    run_importer__dask(models, reimporting=True)

    # the name should not have changed during reimporting
    assert models.Activity.objects.filter(name="Foo").count() == 1
    # the distance however should have been reverted to the original value
    assert models.Activity.objects.filter(distance=original_distance).count() == 1



# Generated by Django 3.2 on 2021-06-11 05:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wkz", "0007_auto_20210430_1633"),
    ]

    operations = [
        migrations.AddField(
            model_name="traces",
            name="total_ascent",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="traces",
            name="total_descent",
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



import datetime
import json
import logging

log = logging.getLogger(__name__)


def migrate_traces(source_model, target_model):
    log.debug(f"migrating models: {source_model} -> {target_model}")
    source = source_model.objects.all()
    log.debug(f"got source query set: {source}")
    for trace in source:
        geojson = trace.geojson
        log.debug(f"got geojson: {geojson}")
        geometry = geojson.replace("'", '"')
        geometry = json.loads(geometry)["features"][0]["geometry"]
        log.debug(f"got geometry: {geometry}")
        target = target_model(
            path_to_file=trace.path_to_file,
            file_name=trace.file_name,
            md5sum=trace.md5sum,
            center_lat=trace.center_lat,
            center_lon=trace.center_lon,
            zoom_level=trace.zoom_level,
            geometry=geometry,
        )
        target.save()


def migrate_duration(activities):
    for activity in activities:
        timedelta = datetime.timedelta(minutes=activity.duration_old)
        log.debug(f"timedelta: {timedelta}")
        activity.duration = timedelta
        activity.save()
        log.debug(
            f"modified activity: {activity.pk}, duration_old: {activity.duration_old}, duration: {activity.duration}"
        )
